#include "mad_dict.h"

/* constants (pre-defined variables) */
char constant_def[] =
"const pi = 4 * atan(1.); "
"const twopi = 2 * pi; "
"const degrad = 180 / pi; "
"const raddeg = pi / 180; "
"const e = exp(1.); "
"const amu0 = 4.e-7 * pi; "
"const emass = 0.510998910e-3; "
"const mumass = 0.1056583668; "
"const nmass = 0.931494028; " /* unified atomic mass unit [GeV/(c**2)] */
"const pmass = 0.938272013; "
"const clight = 299792458.; "
"const qelect = 1.602176487e-19; "
"const hbar = 6.58211899e-25; "
"const erad = 2.8179402894e-15; "
"const prad = erad*emass/pmass; ";

/* mods HG090508 - speed-up twiss
   all new elements must start with at, l, kmax, kmin, calib, polarity
   exactly as the existing ones. Existing elements can only be extended 
   at the end. */

/* command definitions, sorted by module resp. type */

/* The special commands
   "if(", "else{", "elseif(", "while(", ">:macro", ">:line"
   are not included here, but can be found in mad_gcst.c
   under special_comm_desc */
/* IMPORTANT:
   beta0 and twiss MUST have identical portions
   up to "energy" included */

/* format as follows: <...> optional, missing defaults become 0
   "none" stands for empty string
   name : module group type mad_8
   module = the module this command belongs to
   group  = the command group this command belongs to
   type   = 0: default
            1: start of group
            2: end of group
   mad_8  = mad-8 element code
   remark: sequence ... endsequence is NOT a group

   parameter 1 = (type <,default<, call default>>),
   parameter 2 = (type <,default<, call default>>),
   etc.
   some commands contain only data (module "data") for easy access.
   examples:
   l = (r) means: default l = 0, l in command without value -> 0
   energy = (r, 1) means: default energy is 1, in command without value -> 0
   file = (s, none, twiss) means: default no file,
   file in command without value -> twiss

   types: l=logical, i=int, r=real, s=string
   integer and real arrays of arbitrary length can be defined via {...};
   arrays must be defined with maximum length.
*/

/* threader tolerances and maxima: xmax,ymax:  orbit excursion at which
   threader is switched on; att: attenuation factor for kicks.
*/
/* FS 15.03.2004: initial matrix for ptc_twiss, preparation of Touschek */
char command_def[] =
"antiproton: data none 0 0 "  /* comment allowed outside quotes */
"mass     = [r, pmass], "
"charge   = [r, -1]; "
" "
"electron: data none 0 0 "
"mass     = [r, emass], "
"charge   = [r, -1]; "
" "
"negmuon: data none 0 0 "
"mass     = [r, mumass], "
"charge   = [r, -1]; "
" "
"positron: data none 0 0 "
"mass     = [r, emass], "
"charge   = [r, 1]; "
" "
"posmuon: data none 0 0 "
"mass     = [r, mumass], "
"charge   = [r, 1]; "
" "
"proton: data none 0 0 "
"mass     = [r, pmass], "
"charge   = [r, 1]; "
" "
"ion: data none 0 0 "
"mass     = [r, nmass], "
"charge   = [r, 1]; "
" "
"aperture: aperture none 0 0 "
"file     = [s, none, none], "       /* output file with aperture table */
"halofile = [s, none, none], "       /* input file w/halo polygon coordinates */
/* "pipefile = [s, none, none], " */     /* input file w/pipe polygon coordinates */
"range    = [s, #s/#e, none], "      /* range */
"exn      = [r, 3.75e-6], "          /* norm. emit */
"eyn      = [r, 3.75e-6], "          /* norm. emit */
"dqf      = [r, 2.086], "                /* peak linear dispersion */
"betaqfx  = [r, 170.25], "               /* beta x in standard qf */
"dp       = [r, 0.0015], "       /* bucket edge at injection */
"dparx    = [r, 0.273], "        /* fractional parasitic dispersion - horizontal */
"dpary    = [r, 0.273], "        /* fractional parasitic dispersion - vertical */
"cor      = [r, 0.004], "        /* maximum radial closed orbit */
"bbeat    = [r, 1.1], "          /* beta beating coefficient applying to beam size*/
"nco      = [i, 5], "            /* number of azimuth for radial scan */
"halo     = [r, {6., 8.4, 7.3, 7.3}], "  /* halo parameters: n, r, h, v */
"interval = [r, 1.], "           /* length in meters between measurements */
"spec     = [r, 0.], "           /* aperture spec, for plotting only */
"notsimple = [l, false, true], "         /* use if beam pipe is not a simply connex */
"trueprofile = [s, none, none], "    /* read discrete displacements from this file */
"offsetelem = [s, none, none], "     /* read off-axis data from this file */
"refnode = [s, none, none]; "     /* read reference element for the offset file */
" "
"assign: control none 0 0 "
"echo     = [s, terminal, none], "
"truncate = [l, false, true]; "
" "
"beam: control none 0 0 "
"particle = [s, positron, positron], "
"sequence = [s, none, none], "
"bunched  = [l, true, true], "
"radiate  = [l, false, true], "
"mass     = [r, emass], "
"charge   = [r, 1], "
"energy   = [r, 1], "
"pc       = [r, 0], "
"gamma    = [r, 0], "
"ex       = [r, 1], "
"exn      = [r, 0], "
"ey       = [r, 1], "
"eyn      = [r, 0], "
"et       = [r, 1], "
"sigt     = [r, 0], "
"sige     = [r, 0], "
"kbunch   = [r, 1], "
"npart    = [r, 1], "
"bcurrent = [r, 0], "
"freq0    = [r, 0], "
"circ     = [r, 0], "
"dtbyds   = [r, 0], "
"deltap   = [r, 0], "
"beta     = [r, 0], "
"alfa     = [r, 0], "
"u0       = [r, 0], "
"qs       = [r, 0], "
"arad     = [r, 0], "
"bv       = [r, 1], "
"pdamp    = [r, {1,1,2}, {0,0,0}], "
"n1min    = [r, -1]; "
" "
"beta0: control none 0 0 "
"betx     = [r, 0], alfx     = [r, 0], mux      = [r, 0], "
"bety     = [r, 0], alfy     = [r, 0], muy      = [r, 0], "
"x        = [r, 0], px       = [r, 0],         "
"y        = [r, 0], py       = [r, 0],         "
"t        = [r, 0], pt       = [r, 0],         "
"dx       = [r, 0], dpx      = [r, 0],         "
"dy       = [r, 0], dpy      = [r, 0],         "
"wx       = [r, 0], phix     = [r, 0], dmux     = [r, 0], "
"wy       = [r, 0], phiy     = [r, 0], dmuy     = [r, 0], "
"ddx      = [r, 0], ddpx     = [r, 0], "
"ddy      = [r, 0], ddpy     = [r, 0], "
"r11      = [r, 0], r12      = [r, 0], "
"r21      = [r, 0], r22      = [r, 0], "
"energy   = [r, 0]; "
" "
"call: control none 0 0 "
"file = [s, none]; "
" "
"coguess: control control 0 0 "
"x        = [r, 0], "
"px       = [r, 0], "
"y        = [r, 0], "
"py       = [r, 0], "
"t        = [r, 0], "
"pt       = [r, 0], "
"tolerance= [r, 1.0e-6]; "
" "
"create: control none 0 0 "
"table = [s, none], "
"column   = [s,{none}]; "
" "
"delete: control none 0 0 "
"sequence = [s, none], "
"table = [s, none]; "
" "
"deselect: control none 0 0 "
"flag     = [s, none, none], "
"range    = [s, #s/#e, none], "
"class    = [s, none, none], "
"pattern  = [s, any, none], "
"clear    = [l, false, true]; "
" "
"dumpsequ: control none 0 0 "
"sequence = [s, none], "
"level    = [i, 0], "
"file = [s, 0]; "
" "
"exec: control none 0 0; "
" "
"exit: control none 0 0; "
" "
"extract: control none 0 0 "
"sequence = [s, none, none], "
"newname  = [s, none, none], "
"refpos   = [s, none, none], "
"to       = [s, none, none], "
"from     = [s, none, none]; "
" "
"fill: control none 0 0 "
"table = [s, none], "
"row = [i, -1]; "
" "
"setvars: control none 0 0 "
"table = [s, none], "
"row = [i, -1]; "
" "
"help: control none 0 0 "
"dummy    = [s, 0]; "
" "
"option: control none 0 0 "
"bborbit  = [l, false, true], "
"echo     = [l, true, true], "
"info     = [l, true, true], "
"reset    = [l, false, true], "
"debug    = [l, false, true], "
"verbose  = [l, false, true], " /*hbu */
"rbarc    = [l, true, true], "
"thin_foc = [l, true, true], "
"sympl    = [l, true, true], "
"tell     = [l, false, true], "
"threader = [l, false, true], "
"trace    = [l, false, true], "
"verify   = [l, false, true], "
"no_fatal_stop= [l, false, true], "
"freemematstop= [l, false, true], "
"warn     = [l, true, true]; "
" "
"plot: control none 0 0 "
"vaxis    = [s, {none}], "
"vaxis1   = [s, {none}], "
"vaxis2   = [s, {none}], "
"vaxis3   = [s, {none}], "
"vaxis4   = [s, {none}], "
"haxis    = [s, none], "
"hmin     = [r, 0, 0], "
"hmax     = [r, 0, 0], "
"vmin     = [r, {0}], "
"vmax     = [r, {0}], "
"bars     = [i, 0, 1], "
"style    = [i, 1, 1], "
"colour   = [i, 0, 100], "
"symbol   = [i, 0, 1], "
"interpolate   = [l, false, true], "
"spline   = [l, false, true], "
"noline   = [l, false, true], "
"notitle  = [l, false, true], "
"noversion  = [l, false, true], "
"nolegend  = [l, false, true], "
"particle = [s, {none}], "
"multiple = [l, false, true], "
"ptc  = [l, false, true], "
"marker_plot = [l, false, true], "
"range_plot = [l, false, true], "
"zero_suppr = [l, false, true], " /*hbu */
"table    = [s, twiss], "
"ptc_table    = [s, ptc_twiss], "
"title    = [s, none], "
"param    = [s, none], "
"range    = [s, #s/#e, none], "
"file     = [s, none], "
"trackfile  = [s, track]; "
" "
"print: control none 0 0 "
"text     = [s, none]; "
" "
"quit: control none 0 0; "
" "
"readtable: control none 0 0 "
"table = [s, none], "
"file = [s, none]; "
" "
"readmytable: correct correct 0 0 "
"table = [s, none], "
"file  = [s, none]; "
" "
"removefile: control none 0 0 "
"file = [s, none]; "
" "
"renamefile: control none 0 0 "
"file = [s, none], "
"name = [s, none]; "
" "
"resbeam: control none 0 0, "
"sequence = [s, none, none]; "
" "
"resplot: control none 0 0; "
" "
"return: control none 0 0; "
" "
"save: control none 0 0 "
"sequence = [s, {none}], "
"bare     = [l, false, true],"
"beam     = [l, false, true],"
"mad8     = [l, false, true], "
"file     = [s, save, save], "
"newname  = [s, dummy, dummy]; "
" "
"savebeta: control none 0 0 "
"label    = [s, none], "
"place    = [s, none], "
"sequence = [s, none]; "
" "
"select: control none 0 0 "
"flag     = [s, none, none], "
"range    = [s, #s/#e, none], "
"class    = [s, none, none], "
"pattern  = [s, any, none], "
"slice    = [i, 1, 1], "
"thick    = [l, false, true], "
"column   = [s,{none}], "
"sequence = [s, none], "
"full     = [l, false, true], "
"clear    = [l, false, true]; "
" "
"set: control none 0 0 "
"sequence = [s, none], "
"format   = [s,{\"10d\",\"18.10g\",\"-18s\"}]; "
" "
"setplot: control none 0 0 "
"font     = [i, 1], "
"lwidth   = [r, 5], "
"xsize    = [r, 27.], "
"ysize    = [r, 19.], "
"ascale   = [r, 1.5], "
"lscale   = [r, 2], "
"sscale   = [r, 2], "
"rscale   = [r, 1.8], "
"post     = [i, 1], "
"interpolate   = [l, false, true]; "
" "
"sodd: control none 0 0 "
"detune  = [l, false, true], "
"distort1  = [l, false, true], "
"distort2  = [l, false, true], "
"start_stop = [r, {0.0,0.0}], "
"noprint = [l, false, true], "
"nosixtrack  = [l, false, true], "
"print_at_end = [l, false, true], "
"print_all = [l, false, true], "
"multipole_order_range = [i,{1,2}]; "
" "
"show: control none 0 0; "
" "
"stop: control none 0 0; "
" "
"system: control none 0 0 "
"dummy    = [s, 0]; "
" "
"title: control none 0 0 "
"dummy    = [s, 0]; "
" "
"threader: control none 0 0 "
"vector= [r, {5.e-3,5.e-3,1}]; " /* xmax,ymax,att */
" "
"use: control none 0 0 "
"period   = [s, none, none], "
"sequence = [s, none, none], "
"range    = [s, #s/#e, none], "
"survey   = [l, false, true], "
"survtest = [l, false, true]; "
" "
"value: control none 0 0 "
"dummy    = [s, 0]; "
" "
"write: control none 0 0 "
"table = [s, none], "
"file = [s, none]; "
" "
"sixtrack: c6t none 0 0 "
"mult_auto_off = [l, false, true], "
"cavall        = [l, false, true], "
"aperture      = [l, false, true], "
"split         = [l, false, true], "
"max_mult_ord  = [i, 11], "
"radius        = [r, 1]; "
" "
"ptc_create_universe: ptc_create_universe none 0 0 "
"sector_nmul_max = [i, 10], "
"sector_nmul = [i, 5], "
"symprint = [l, true], "
"ntpsa = [l, false, true]; "
" "
"ptc_export_xml: ptc_export_xml none 0 0 " /* space after 0! */
"file = [s,output.xml,output.xml] ; " /* defaults to output.xml */
" "
"ptc_create_layout: ptc_create_layout none 0 0 "
"time          = [l, true, true], "
"model         = [i, 1], "
"method        = [i, 2], "
"exact         = [l, false, true], "
"closed_layout = [l, true, true], "
"nst           = [i, 1], "
"mad_mult      = [l, true], "
"mad8          = [l, true], "
"offset_deltap = [r, 0], "
"errors_out    = [l, false, true], "
"magnet_name   = [s, none, none], "
"resplit       = [l, false, true], "
"thin          = [r, 0.001],  "
"xbend         = [r, -1, 0.001],  "
"even          = [l, true, true]; "
" "
"ptc_read_errors: ptc_read_errors none 0 0 "
"overwrite     = [l, false, true]; "
" "
"ptc_refresh_k: ptc_refresh_k none 0 0 "
"overwrite     = [l, false, true]; "
" "
"ptc_move_to_layout: ptc_move_to_layout none 0 0 "
"index = [i, 1]; "
" "
"ptc_align: ptc_align none 0 0; "
" "
"ptc_select: ptc_select none 0 0 "
"table      = [s, none, none], "
"column     = [s, none, none], "
"polynomial = [i, none] , "/*defines the element of the tracked 6D array of polynomials */
"monomial   = [s, none] , "/*defines the monomial that coefficient will be sent to table */
"parametric = [l, false, true], "/*Tells if the result should be stored in a parametric form if knobs are present*/
"quantity   = [s, {none}] ; "
" "
"ptc_select_moment: ptc_select_moment none 0 0 "
"table      = [s, moments, none], "
"column     = [s, none, none], "
"parametric = [l, false, true], "/*Tells if the result should be stored in a parametric form if knobs are present*/
/*"place    = [s, none], "*/
"moment_s   = [s, {none}], "
"moment     = [i, {0}], "
"quantity   = [s, {none}] ; "
" "
"ptc_knob: ptc_knob none 0 0 "
"initial = [s, none] , "/* */
"element = [s, none] , "/* */
"kn    = [i, {-1}], "
"ks    = [i, {-1}], "
"exactmatch = [l, true, true]; "
" "
"ptc_putbeambeam: ptc_putbeambeam none 0 0 "
"element = [s, none] , "/* */
"local_s = [r, -1] , "/* */
"global_s = [r, -1] , "/* */
"xma = [r, none] , "/* */
"yma = [r, none] , "/* */
"sigx  = [r, none], "
"sigy  = [r, none]; "
" "
"ptc_varyknob: ptc_varyknob none 0 0 "
"initial = [s, none] , "/* */
"element = [s, none] , "/* */
"kn    = [i, -1], "
"ks    = [i, -1], "
"exactmatch = [l, true, true], "
"trustrange    = [r, 0.1], " /*defines the range the expansion is trusted*/
"step     = [r, 0.0], "
"lower    = [r, -1.e20], "
"upper    = [r,  1.e20]; "
" "
"ptc_setknobvalue: ptc_setknobvalue none 0 0 "
"element = [s, none] , "/* */
"kn    = [i, -1], "
"ks    = [i, -1], "
"refreshtables  = [l, true, true], "
"value = [r] ; "
" "
"ptc_refreshpartables: ptc_refreshpartables none 0 0 "
"refreshtables  = [l, true, true] ;"
" "
"rviewer: rviewer none 0 0 "
"exactmatch = [l, true, true] ; "
" "
"ptc_printparametric: ptc_printparametric none 0 0 "
"filename = [s, none] , "/* */
"format = [s, ptc] ; "
" "
"ptc_setfieldcomp: ptc_setfieldcomp none 0 0 "
"fromerrtable    = [s, none, efield], "
"add             = [l, false, true], "
"element = [s, none] , "/* */
"kn    = [i, -1], "
"ks    = [i, -1], "
"value = [r] ; "
" "
"ptc_eplacement: ptc_eplacement none 0 0 "
"range = [s, none] , "/* */
"x        = [r, 0], y       = [r, 0], z       = [r, 0], " /*position of the front face*/
"phi      = [r, 0], " /**/
"theta   = [r, 0], "
"surveyall    = [l, true, true] , "
"onlyposition    = [l, false, true] , "/* if true changes only position and, leaves orinetation untouched */
"onlyorientation = [l, false, true] , "/* if true changes only orientation and, leaves position untouched */
"autoplacedownstream = [l, true, true] , "/* if true all the elements downstream are placed at default positions, if false the rest of the layout stays antouched  */
"refframe = [s, gcs] ; " /* coordinate system, gcs - global coordinate syste, current - current position, previouselement */
" "
"ptc_printframes: ptc_printframes none 0 0 "
"file = [s, none] , "/* */
"with = [s, none] , "/* */
"format = [s, text] ; "
" "
"ptc_twiss: ptc_twiss none 0 0 "
"betx     = [r, 0], alfx     = [r, 0], mux      = [r, 0], "
"bety     = [r, 0], alfy     = [r, 0], muy      = [r, 0], "
"x        = [r, 0], px       = [r, 0],         "
"y        = [r, 0], py       = [r, 0],         "
"t        = [r, 0], pt       = [r, 0],         "
"dx       = [r, 0], dpx      = [r, 0],         "
"dy       = [r, 0], dpy      = [r, 0],         "
"wx       = [r, 0], phix     = [r, 0], dmux     = [r, 0], "
"wy       = [r, 0], phiy     = [r, 0], dmuy     = [r, 0], "
"ddx      = [r, 0], ddpx     = [r, 0], "
"ddy      = [r, 0], ddpy     = [r, 0], "
"r11      = [r, 0], r12      = [r, 0], "
"r21      = [r, 0], r22      = [r, 0], "
"energy   = [r, 0], "
"file     = [s, none, ptc_twiss], "
"table    = [s, none, ptc_twiss], "
"summary_file  = [s, none, ptc_twiss_summary], "
"summary_table = [s, none, ptc_twiss_summary], "
"savemaps = [l, true, true], " /*buffers maps for future usage as universal_taylors */
"writetmap = [l,false, true], " /*writes transfer.map */
"slice_magnets  = [l,false,true], "
"center_magnets = [l,false,true], "
"deltap_dependency = [l,false,true], " /* compute Twiss derivatives w.r.t deltap*/
"xdistr   = [s, gauss, gauss], " /* type of distribution in X plane*/
"ydistr   = [s, gauss, gauss], " /* type of distribution in Y plane*/
"zdistr   = [s, gauss, gauss], " /* type of distribution in Z plane*/
"deltapreadsize = [r, 0], " /*size on delta */
"bunchlegngth = [r, 0], " /*size in Z/T*/
"initial_moments_manual = [l, false, true], " /* re?? are moments at beginning of the line, 21 in case of 3DGauss, or 10 in fully coupled */
"deltap = [r, 0], "
"range    = [s, #s/#e, none], "
"icase = [i, 4], "
"exact_mis = [l, false, true], "
"time = [l, false, true], "
"no = [i, 1], "
"closed_orbit = [l, false, true], "
"rmatrix  = [l, false, true], "
"initial_matrix_table = [l, false, true], "
"initial_matrix_manual = [l, false, true], "
"initial_map_manual = [l, false, true], "
"initial_ascript_manual = [l, false, true], "
"re11     = [r, 1], re12     = [r, 0], re13     = [r, 0], "
"re14     = [r, 0], re15     = [r, 0], re16     = [r, 0], "
"re21     = [r, 0], re22     = [r, 1], re23     = [r, 0], "
"re24     = [r, 0], re25     = [r, 0], re26     = [r, 0], "
"re31     = [r, 0], re32     = [r, 0], re33     = [r, 1], "
"re34     = [r, 0], re35     = [r, 0], re36     = [r, 0], "
"re41     = [r, 0], re42     = [r, 0], re43     = [r, 0], "
"re44     = [r, 1], re45     = [r, 0], re46     = [r, 0], "
"re51     = [r, 0], re52     = [r, 0], re53     = [r, 0], "
"re54     = [r, 0], re55     = [r, 1], re56     = [r, 0], "
"re61     = [r, 0], re62     = [r, 0], re63     = [r, 0], "
"re64     = [r, 0], re65     = [r, 0], re66     = [r, 1], "
"betz     = [r, 0], alfz     = [r, 0], muz      = [r, 0], "
"beta0    = [s, none, beta0]; "
" "
"ptc_moments: ptc_moments none 0 0 "
"no = [i, 1], "
"xdistr   = [s, gauss, gauss], " /* type of distribution in X plane*/
"ydistr   = [s, gauss, gauss], " /* type of distribution in Y plane*/
"zdistr   = [s, gauss, gauss]; " /* type of distribution in Z plane*/
" "
"select_ptc_normal: select_ptc_normal none 0 0 "
"dx =   [s, {none}], "
"dpx =  [s, {none}], "
"dy =   [s, {none}], "
"dpy =  [s, {none}], "
"q1 =   [s, {none}], "
"dq1 =  [s, {none}], "
"q2 =   [s, {none}], "
"dq2 =  [s, {none}], "
"gnfu = [s, {none}], "
"haml = [s, {none}], "
"anhx = [s, {none}], "
"anhy = [s, {none}], "
"eign = [s, {none}], "
"clear = [l, false, true]; "
" "
"ptc_normal: ptc_normal none 0 0 "
"icase = [i, 4], "
"no = [i, 1], "
"normal = [l, false, true], "
"closed_orbit = [l, false, true], "
"maptable = [l, false, true], "
"exact_mis = [l, false, true], "
"time = [l, false, true], "
"deltap = [r, 0], "
"x        = [r, 0], "
"px       = [r, 0], "
"y        = [r, 0], "
"py       = [r, 0], "
"t        = [r, 0], "
"pt       = [r, 0]; "
" "                               /*Piotr.Skowronski@cern.ch*/
"ptc_trackline: ptc_trackline none 0 0 " /*implemented by subroutine ptc_trackline() in madx_ptc_trackline.f90*/
"turns        = [i, 1], "
"ffile    = [i, 1], "
"file     = [s, track, track], "
"gcs   = [l, false, true], "
"everystep   = [l, false, true], "
"tableallsteps  = [l, false, true], "
"extension= [s, none, none], "
"rootntuple = [l, false, true], "/*writes all the tracks in ROOT NTuple -> works only if program is dynamically linked and RPLOT plugin is present */
"onetable = [l, false, true]; "
" "                               /*Piotr.Skowronski@cern.ch*/
"ptc_dumpmaps: ptc_dumpmaps none 0 0 " /*implemented by subroutine ptc_dumpmaps() in madx_ptc_module.f90*/
"range    = [s,  #s/#e], "
"frombegin     = [s, ptcmaps, ptcmaps], "
"file     = [s, ptcmap, ptcmaps]; "
" "
"ptc_oneturnmap: ptc_oneturnmap none 0 0 " /*implemented by subroutine ptc_dumpmaps() in madx_ptc_module.f90*/
"x        = [r, 0], "
"px       = [r, 0], "
"y        = [r, 0], "
"py       = [r, 0], "
"t        = [r, 0], "
"pt       = [r, 0], "
"range    = [s,  #s/#e], "
"file     = [s, ptcmap, ptcmaps]; "
" "
"ptc_track_shape: ptc_track_shape none 0 0 "
"debug = [i, 0, 1]; "
" "
"ptc_setswitch: ptc_setswitch none 0 0 "
"debuglevel = [i,1], "/*sets the level of debugging printout 0 none, 4 everything */
"maxacceleration = [l, true, true], " /*switch saying to set cavities phases so the reference orbit is always on the crest, i.e. gains max energy*/
"exact_mis = [l, false, true], " /* switch to ensure exact misaligment treatment */
"totalpath = [l, false, true], "      /* switch to use totalpath, modifies PTC states by adding totalpath0 flag */
"radiation = [l, false, true], "  /*sets the radiation switch/internal state of PTC */
"fringe = [l, false, true], "  /*sets the fringe switch/internal state of PTC */
"nocavity = [l, true, true], "  /*sets the nocavity switch/internal state of PTC */
"time = [l, true, true]; "      /* switch to use time rather than pathlength, modifies PTC states by adding TIME0 flag */
" "
"ptc_enforce6d: ptc_enforce6d none 0 0 "
"flag = [l, false, false]; " /* normally 6D is reduced to 4D if no cavities are present. This switch prevents it. It is needed to calcualte  fg R56 in a chicane */
" "
"ptc_track: ptc_track none 0 0 "
"icase        = [i, 4], "
"turns        = [i, 1], "
"closed_orbit = [l, false, true], "
"deltap       = [r, 0], "
"exact_mis = [l, false, true], "
"time = [l, false, true], "
"element_by_element = [l, false, true], "
"radiation = [l, false, true], "
"radiation_model1 = [l, false, true], "
"radiation_energy_loss = [l, false, true], "
"radiation_quad = [l, false, true], "
"space_charge = [l, false, true], "
"beam_envelope = [l, false, true], "
"dump     = [l, false, true], "
"onetable = [l, false, true], "
"file     = [s, track, track], "
"extension= [s, none, none], "
"ffile    = [i, 1], "
"maxaper= [r, {0.1, 0.01, 0.1, 0.01, 1., 0.1}], "
"norm_out = [l, false, true], "
"norm_no  = [i, 1]; "
" "
"ptc_start: ptc_start none 0 0 "
"x        = [r, 0], "
"px       = [r, 0], "
"y        = [r, 0], "
"py       = [r, 0], "
"t        = [r, 0], "
"pt       = [r, 0], "
"fx       = [r, 0], "
"phix     = [r, 0], "
"fy       = [r, 0], "
"phiy     = [r, 0], "
"ft       = [r, 0], "
"phit     = [r, 0]; "
" "
"ptc_observe: ptc_observe none 0 0 "
"place    = [s, none]; "
" "
"ptc_track_end: ptc_track_end none 0 0; "
" "                               /*Piotr.Skowronski@cern.ch*/
"ptc_script: ptc_script none 0 0 " /*implemented by subroutine ptc_dumpmaps() in madx_ptc_script.f90*/
"file     = [s, script.ptc, script.ptc]; "
" "
"ptc_open_gino: ptc_open_gino 0 0 "
"command     = [s, dothis, dothis]; "
" "
"ptc_end: ptc_end none 0 0; "
" "
"correct: correct correct 0 0 "
"error    = [r, 1.e-5], "
"ncorr    = [i, 0], "
"monerror = [i, 0], "
"monscale = [i, 0], "
"monon    = [r, 1], "
"moncut   = [r, 0], "
"corrlim  = [r, 1], "
"resout   = [i, 0], "
"sequence = [s, {none}], "
"clist    = [s, none], "
"mlist    = [s, none], "
"plane    = [s, x], "
"flag     = [s, ring], "
"mode     = [s, micado], "
"cond     = [i, 0], "
"twissum  = [i, 0], "
"orbit    = [s, none], "
"model    = [s, none], "
"corzero  = [i, 0], "
"tworing  = [l, false, true], "
"beam1tab = [s, none], "
"beam2tab = [s, none], "
"target   = [s, none], "
"name_col = [s, none], "
"x_col    = [s, none], "
"y_col    = [s, none], "
"units    = [r, 1], "
"sngcut   = [r, 50], "
"sngval   = [r, 2], "
"extern   = [l, false, true]; "
" "
"getorbit: correct correct 0 0 "
"file = [s, orbit, orbit]; "
" "
"readcorr: correct correct 0 0 "
"file = [s, orbit, orbit]; "
" "
"setcorr: correct correct 0 0 "
"table = [s, none], "
"file = [s, orbit, orbit]; "
" "
"putorbit: correct correct 0 0 "
"file = [s, orbit, orbit]; "
" "
"getkick: correct correct 0 0 "
"file = [s, setting, setting], "
"plane    = [s, none, none], "
"add      = [l, false], "
"direct   = [l, false]; "
" "
"putkick: correct correct 0 0 "
"file = [s, setting, setting], "
"plane    = [s, none, none], "
"direct   = [l, false]; "
" "
"getdisp: correct correct 0 0 "
"file = [s, dispersion, dispersion]; "
" "
"putdisp: correct correct 0 0 "
"file = [s, dispersion, dispersion]; "
" "
"usekick: correct correct 0 0 "
"sequence = [s, none, none], "
"status   = [s, on], "
"range    = [s,  #s/#e], "
"class    = [s, none], "
"pattern  = [s, any, none]; "
" "
"usemonitor: correct correct 0 0 "
"sequence = [s, none, none], "
"status   = [s, on], "
"range    = [s,  #s/#e], "
"class    = [s, none], "
"pattern  = [s, any, none]; "
" "
"coption: correct none 0 0 "
"seed     = [i, 123456789], "
"print    = [i, 1],  "
"debug    = [i, 0];  "
" "
"seqedit: edit edit 1 0 "
"sequence = [s, none, none]; "
" "
"flatten: edit edit 0 0; "
" "
"install: edit edit 0 0  "
"element  = [s, none, none], "
"class    = [s, none, none], "
"at       = [r, 0, 0], "
"from     = [s, none, none]; "
" "
"move: edit edit 0 0 "
"element  = [s, none, none], "
"by       = [r, 0, 0], "
"to       = [r, 0, 0], "
"from     = [s, none, none]; "
" "
"remove: edit edit 0 0 "
"element  = [s, none, none]; "
" "
"cycle: edit edit 0 0 "
"start    = [s, none, none]; "
" "
"reflect: edit edit 0 0; "
" "
"replace: edit edit 0 0 "
"element  = [s, none, none], "
"by       = [s, none, none]; "
" "
"endedit: edit edit 2 0; "
" "
"drift: element none 0 1 "
"at       = [r, 1.e20], "
"l        = [r, 0], "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"slice    = [i, 1],  "
"magnet   = [i, 0], "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"from     = [s, none], "
"type     = [s, none, none], "
"rbend_length    = [s, none], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true]; "
" "
"rbend: element none 0 2 "
"at       = [r, 1.e20], "
"l        = [r, 0],  "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"angle    = [r, 0],  "
"tilt     = [r, 0],  "
"k0       = [r, 0],  "
"k0s      = [r, 0],  "
"k1       = [r, 0],  "
"k1s      = [r, 0],  "
"e1       = [r, 0],  "
"e2       = [r, 0],  "
"k2       = [r, 0],  "
"k2s      = [r, 0],  "
"h1       = [r, 0],  "
"h2       = [r, 0],  "
"hgap     = [r, 0],  "
"fint     = [r, 0, 0.5],  "
"fintx    = [r, -1.0], "
"k3       = [r, 0],  "
"k3s      = [r, 0],  "
"fcsr     = [r, 0],  "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"aperture = [r, {0}], "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"knl      = [r, {0}], "
"ksl      = [r, {0}], "
"magnet   = [i, 1],  "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"slice    = [i, 1],  "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"apertype = [s, circle, circle], "
"from     = [s, none], "
"type     = [s, none, none], "
"ptcrbend        = [l, false, true], "
"truerbend       = [l, true, true], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, true, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true], "
"add_angle       = [r, {0,0,0,0,0}]; "
" "
"sbend: element none 0 3 "
"at       = [r, 1.e20], "
"l        = [r, 0],  "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"angle    = [r, 0],  "
"tilt     = [r, 0],  "
"k0       = [r, 0],  "
"k0s      = [r, 0],  "
"k1       = [r, 0],  "
"k1s      = [r, 0],  "
"e1       = [r, 0],  "
"e2       = [r, 0],  "
"k2       = [r, 0],  "
"k2s      = [r, 0],  "
"h1       = [r, 0],  "
"h2       = [r, 0],  "
"hgap     = [r, 0],  "
"fint     = [r, 0, 0.5],  "
"fintx    = [r, -1.0], "
"k3       = [r, 0],  "
"k3s      = [r, 0],  "
"fcsr     = [r, 0],  "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"aperture = [r, {0}], "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"knl      = [r, {0}], "
"ksl      = [r, {0}], "
"magnet   = [i, 1],  "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"slice    = [i, 1],  "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"apertype = [s, circle, circle], "
"from     = [s, none], "
"type     = [s, none, none], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, true, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true]; "
" "
"matrix: element none 0 4 "
"at       = [r, 1.e20], "
"l        = [r, 0],  "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"kick1    = [r, 0], kick2    = [r, 0], kick3    = [r, 0], "
"kick4    = [r, 0], kick5    = [r, 0], kick6    = [r, 0], "
"rm11     = [r, 1], rm12     = [r, 0], rm13     = [r, 0], "
"rm14     = [r, 0], rm15     = [r, 0], rm16     = [r, 0], "
"rm21     = [r, 0], rm22     = [r, 1], rm23     = [r, 0], "
"rm24     = [r, 0], rm25     = [r, 0], rm26     = [r, 0], "
"rm31     = [r, 0], rm32     = [r, 0], rm33     = [r, 1], "
"rm34     = [r, 0], rm35     = [r, 0], rm36     = [r, 0], "
"rm41     = [r, 0], rm42     = [r, 0], rm43     = [r, 0], "
"rm44     = [r, 1], rm45     = [r, 0], rm46     = [r, 0], "
"rm51     = [r, 0], rm52     = [r, 0], rm53     = [r, 0], "
"rm54     = [r, 0], rm55     = [r, 1], rm56     = [r, 0], "
"rm61     = [r, 0], rm62     = [r, 0], rm63     = [r, 0], "
"rm64     = [r, 0], rm65     = [r, 0], rm66     = [r, 1], "
"tm111     = [r, 0], tm112     = [r, 0], tm113     = [r, 0], "
"tm114     = [r, 0], tm115     = [r, 0], tm116     = [r, 0], "
"tm121     = [r, 0], tm122     = [r, 0], tm123     = [r, 0], "
"tm124     = [r, 0], tm125     = [r, 0], tm126     = [r, 0], "
"tm131     = [r, 0], tm132     = [r, 0], tm133     = [r, 0], "
"tm134     = [r, 0], tm135     = [r, 0], tm136     = [r, 0], "
"tm141     = [r, 0], tm142     = [r, 0], tm143     = [r, 0], "
"tm144     = [r, 0], tm145     = [r, 0], tm146     = [r, 0], "
"tm151     = [r, 0], tm152     = [r, 0], tm153     = [r, 0], "
"tm154     = [r, 0], tm155     = [r, 0], tm156     = [r, 0], "
"tm161     = [r, 0], tm162     = [r, 0], tm163     = [r, 0], "
"tm164     = [r, 0], tm165     = [r, 0], tm166     = [r, 0], "
"tm211     = [r, 0], tm212     = [r, 0], tm213     = [r, 0], "
"tm214     = [r, 0], tm215     = [r, 0], tm216     = [r, 0], "
"tm221     = [r, 0], tm222     = [r, 0], tm223     = [r, 0], "
"tm224     = [r, 0], tm225     = [r, 0], tm226     = [r, 0], "
"tm231     = [r, 0], tm232     = [r, 0], tm233     = [r, 0], "
"tm234     = [r, 0], tm235     = [r, 0], tm236     = [r, 0], "
"tm241     = [r, 0], tm242     = [r, 0], tm243     = [r, 0], "
"tm244     = [r, 0], tm245     = [r, 0], tm246     = [r, 0], "
"tm251     = [r, 0], tm252     = [r, 0], tm253     = [r, 0], "
"tm254     = [r, 0], tm255     = [r, 0], tm256     = [r, 0], "
"tm261     = [r, 0], tm262     = [r, 0], tm263     = [r, 0], "
"tm264     = [r, 0], tm265     = [r, 0], tm266     = [r, 0], "
"tm311     = [r, 0], tm312     = [r, 0], tm313     = [r, 0], "
"tm314     = [r, 0], tm315     = [r, 0], tm316     = [r, 0], "
"tm321     = [r, 0], tm322     = [r, 0], tm323     = [r, 0], "
"tm324     = [r, 0], tm325     = [r, 0], tm326     = [r, 0], "
"tm331     = [r, 0], tm332     = [r, 0], tm333     = [r, 0], "
"tm334     = [r, 0], tm335     = [r, 0], tm336     = [r, 0], "
"tm341     = [r, 0], tm342     = [r, 0], tm343     = [r, 0], "
"tm344     = [r, 0], tm345     = [r, 0], tm346     = [r, 0], "
"tm351     = [r, 0], tm352     = [r, 0], tm353     = [r, 0], "
"tm354     = [r, 0], tm355     = [r, 0], tm356     = [r, 0], "
"tm361     = [r, 0], tm362     = [r, 0], tm363     = [r, 0], "
"tm364     = [r, 0], tm365     = [r, 0], tm366     = [r, 0], "
"tm411     = [r, 0], tm412     = [r, 0], tm413     = [r, 0], "
"tm414     = [r, 0], tm415     = [r, 0], tm416     = [r, 0], "
"tm421     = [r, 0], tm422     = [r, 0], tm423     = [r, 0], "
"tm424     = [r, 0], tm425     = [r, 0], tm426     = [r, 0], "
"tm431     = [r, 0], tm432     = [r, 0], tm433     = [r, 0], "
"tm434     = [r, 0], tm435     = [r, 0], tm436     = [r, 0], "
"tm441     = [r, 0], tm442     = [r, 0], tm443     = [r, 0], "
"tm444     = [r, 0], tm445     = [r, 0], tm446     = [r, 0], "
"tm451     = [r, 0], tm452     = [r, 0], tm453     = [r, 0], "
"tm454     = [r, 0], tm455     = [r, 0], tm456     = [r, 0], "
"tm461     = [r, 0], tm462     = [r, 0], tm463     = [r, 0], "
"tm464     = [r, 0], tm465     = [r, 0], tm466     = [r, 0], "
"tm511     = [r, 0], tm512     = [r, 0], tm513     = [r, 0], "
"tm514     = [r, 0], tm515     = [r, 0], tm516     = [r, 0], "
"tm521     = [r, 0], tm522     = [r, 0], tm523     = [r, 0], "
"tm524     = [r, 0], tm525     = [r, 0], tm526     = [r, 0], "
"tm531     = [r, 0], tm532     = [r, 0], tm533     = [r, 0], "
"tm534     = [r, 0], tm535     = [r, 0], tm536     = [r, 0], "
"tm541     = [r, 0], tm542     = [r, 0], tm543     = [r, 0], "
"tm544     = [r, 0], tm545     = [r, 0], tm546     = [r, 0], "
"tm551     = [r, 0], tm552     = [r, 0], tm553     = [r, 0], "
"tm554     = [r, 0], tm555     = [r, 0], tm556     = [r, 0], "
"tm561     = [r, 0], tm562     = [r, 0], tm563     = [r, 0], "
"tm564     = [r, 0], tm565     = [r, 0], tm566     = [r, 0], "
"tm611     = [r, 0], tm612     = [r, 0], tm613     = [r, 0], "
"tm614     = [r, 0], tm615     = [r, 0], tm616     = [r, 0], "
"tm621     = [r, 0], tm622     = [r, 0], tm623     = [r, 0], "
"tm624     = [r, 0], tm625     = [r, 0], tm626     = [r, 0], "
"tm631     = [r, 0], tm632     = [r, 0], tm633     = [r, 0], "
"tm634     = [r, 0], tm635     = [r, 0], tm636     = [r, 0], "
"tm641     = [r, 0], tm642     = [r, 0], tm643     = [r, 0], "
"tm644     = [r, 0], tm645     = [r, 0], tm646     = [r, 0], "
"tm651     = [r, 0], tm652     = [r, 0], tm653     = [r, 0], "
"tm654     = [r, 0], tm655     = [r, 0], tm656     = [r, 0], "
"tm661     = [r, 0], tm662     = [r, 0], tm663     = [r, 0], "
"tm664     = [r, 0], tm665     = [r, 0], tm666     = [r, 0], "
"from     = [s, none], "
"type     = [s, none, none], "
"magnet   = [i, 1],  "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true]; "
" "
"quadrupole: element none 0 5 "
"at       = [r, 1.e20], "
"l        = [r, 0],  "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"tilt     = [r, 0],  "
"k1       = [r, 0],  "
"k1s      = [r, 0],  "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"aperture = [r, {0}], "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"knl      = [r, {0}], "
"ksl      = [r, {0}], "
"slice    = [i, 1],  "
"thick    = [l, false, true], "
"magnet   = [i, 1],  "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"from     = [s, none], "
"type     = [s, none, none], "
"apertype = [s, circle, circle], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true], "
"k0       = [r, 0];  "
" "
"sextupole: element none 0 6 "
"at       = [r, 1.e20], "
"l        = [r, 0],  "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"tilt     = [r, 0],  "
"k2       = [r, 0],  "
"k2s      = [r, 0],  "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"aperture = [r, {0}], "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"knl      = [r, {0}], "
"ksl      = [r, {0}], "
"magnet   = [i, 1],  "
"slice    = [i, 1],  "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"apertype = [s, circle, circle], "
"from     = [s, none], "
"type     = [s, none, none], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true]; "
" "
"octupole: element none 0 7 "
"at       = [r, 1.e20], "
"l        = [r, 0],  "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"tilt     = [r, 0],  "
"k3       = [r, 0],  "
"k3s      = [r, 0],  "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"knl      = [r, {0}], "
"ksl      = [r, {0}], "
"aperture = [r, {0}], "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"slice    = [i, 1],  "
"magnet   = [i, 1],  "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"from     = [s, none], "
"type     = [s, none, none], "
"apertype = [s, circle, circle], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true]; "
" "
"multipole: element none 0 8 "
"at       = [r, 1.e20], "
"l        = [r, 0], "
"noise     = [r, 0], "
"noisemax    = [i, 1],  "
"npeak     = [r, {0}], "
"ntune     = [r, {0}], "
"nlag     = [r, {0}], "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"tilt     = [r, 0],  "
"lrad     = [r, 0], "
"knl      = [r, {0}], "
"ksl      = [r, {0}], "
"apertype = [s, circle, circle], "
"aperture = [r, {0}],  "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"slice    = [i, 1],  "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"magnet   = [i, 1],  "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"from     = [s, none], "
"type     = [s, none, none], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true]; "
" "
"solenoid: element none 0 9 "
"at       = [r, 1.e20], "
"l        = [r, 0],  "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"lrad     = [r, 0],  " /* hbu */
"ks       = [r, 0],  "
"ksi      = [r, 0],  " /* was: ksl, but that clashes with naming conventions of multipoles */
"knl      = [r, {0}], "
"ksl      = [r, {0}], "
"slice    = [i, 1],  "
"apertype = [s, circle, circle], "
"aperture = [r, {0}], "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"magnet   = [i, 1],  "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"from     = [s, none], "
"type     = [s, none, none], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true]; "
" "
"rfcavity: element none 0 10 "
"at       = [r, 1.e20], "
"l        = [r, 0],  "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"volt     = [r, 0],  "
"lag      = [r, 0],  "
"freq     = [r, 0],  "
"betrf    = [r, 0],  "
"pg       = [r, 0],  "
"shunt    = [r, 0],  "
"tfill    = [r, 0],  "
"eloss    = [r, 0],  "
"volterr  = [r, 0],  "
"lagerr   = [r, 0],  "
"binmax   = [r, 0],  "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"aperture = [r, {0}],  "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"harmon   = [i, 0],  "
"nbin     = [i, 0],  "
"magnet   = [i, 0],  "
"slice    = [i, 1],  "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"n_bessel        = [i, 0], "
"from     = [s, none], "
"type     = [s, none, none], "
"lfile    = [s, none, lfile],  "
"tfile    = [s, none, tfile], "
"apertype = [s, circle, circle], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true], "
"no_cavity_totalpath = [l, false, true]; "
" "
"elseparator: element none 0 11 "
"at       = [r, 1.e20], "
"l        = [r, 0],  "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"tilt     = [r, 0],  "
"ex       = [r, 0],  "
"ey       = [r, 0],  "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"aperture = [r, {0}],  "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"slice    = [i, 1], "
"magnet   = [i, 0],  "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"from     = [s, none], "
"type     = [s, none, none], "
"apertype = [s, circle, circle], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true]; "
" "
"srotation: element none 0 12 "
"at       = [r, 1.e20], "
"l        = [r, 0],  "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"apertype = [s, circle, circle], "
"aperture = [r, {0}],  "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"angle    = [r, 0], "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"magnet   = [i, 0],  "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"from     = [s, none], "
"type     = [s, none, none], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true]; "
" "
"yrotation: element none 0 13 "
"at       = [r, 1.e20], "
"l        = [r, 0],  "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"apertype = [s, circle, circle], "
"aperture = [r, {0}],  "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"angle    = [r, 0], "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"magnet   = [i, 0],  "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"from     = [s, none], "
"type     = [s, none, none], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true]; "
" "
"hkicker: element none 0 14 "
"at       = [r, 1.e20], "
"l        = [r, 0],  "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"tilt     = [r, 0],  "
"lrad     = [r, 0],  "
"kick     = [r, 0],  "
"hkick    = [r, 0],  "
"chkick   = [r, 0],  "
"chflag   = [i, 1],  "
"slice    = [i, 1],  "
"sinkick  = [i, 0],  "
"sinpeak  = [r, 0],  "
"sintune  = [r, 0],  "
"sinphase = [r, 0],  "
"apertype = [s, circle, circle], "
"aperture = [r, {0}],  "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"magnet   = [i, 1],  "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"from     = [s, none], "
"type     = [s, none, none], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true]; "
" "
"kicker: element none 0 15 "
"at       = [r, 1.e20], "
"l        = [r, 0],  "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"tilt     = [r, 0],  "
"lrad     = [r, 0],  "
"hkick    = [r, 0],  "
"vkick    = [r, 0],  "
"chkick   = [r, 0],  "
"cvkick   = [r, 0],  "
"chflag   = [i, 1],  "
"cvflag   = [i, 1],  "
"slice    = [i, 1],  "
"sinkick  = [i, 0],  "
"sinpeak  = [r, 0],  "
"sintune  = [r, 0],  "
"sinphase = [r, 0],  "
"apertype = [s, circle, circle], "
"aperture = [r, {0}],  "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"magnet   = [i, 1],  "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"from     = [s, none], "
"type     = [s, none, none], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true]; "
" "
"vkicker: element none 0 16 "
"at       = [r, 1.e20], "
"l        = [r, 0],  "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"tilt     = [r, 0],  "
"lrad     = [r, 0],  "
"kick     = [r, 0],  "
"vkick    = [r, 0],  "
"cvkick   = [r, 0],  "
"cvflag   = [i, 1],  "
"magnet   = [i, 1],  "
"slice    = [i, 1],  "
"sinkick  = [i, 0],  "
"sinpeak  = [r, 0],  "
"sintune  = [r, 0],  "
"sinphase = [r, 0],  "
"from     = [s, none], "
"type     = [s, none, none], "
"apertype = [s, circle, circle], "
"aperture = [r, {0}],  "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true]; "
" "
"hmonitor: element none 0 17 "
"at       = [r, 1.e20], "
"l        = [r, 0], "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"apertype = [s, circle, circle], "
"aperture = [r, {0}],  "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"magnet   = [i, 0],  "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"from     = [s, none], "
"type     = [s, none, none], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true]; "
" "
"monitor: element none 0 18 "
"at       = [r, 1.e20], "
"l        = [r, 0], "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"magnet   = [i, 0],  "
"apertype = [s, circle, circle], "
"aperture = [r, {0}],  "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"from     = [s, none], "
"type     = [s, none, none], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true]; "
" "
"vmonitor: element none 0 19 "
"at       = [r, 1.e20], "
"l        = [r, 0], "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"apertype = [s, circle, circle], "
"aperture = [r, {0}],  "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"magnet   = [i, 0],  "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"from     = [s, none], "
"type     = [s, none, none], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true]; "
" "
"ecollimator: element none 0 20 "
"at       = [r, 1.e20], "
"l        = [r, 0],  "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"tilt     = [r, 0],  "
"lrad     = [r, 0],  "
"magnet   = [i, 0],  "
"apertype = [s, circle, circle], "
"aperture = [r, {0}],  "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"xsize    = [r, 0],  "
"ysize    = [r, 0], "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"from     = [s, none], "
"type     = [s, none, none], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true]; "
" "
"rcollimator: element none 0 21 "
"at       = [r, 1.e20], "
"l        = [r, 0],  "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"tilt     = [r, 0],  "
"lrad     = [r, 0],  "
"magnet   = [i, 0],  "
"apertype = [s, circle, circle], "
"aperture = [r, {0}],  "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"xsize    = [r, 0],  "
"ysize    = [r, 0], "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"from     = [s, none], "
"type     = [s, none, none], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true]; "
" "
"beambeam: element none 0 22 "
"at       = [r, 1.e20], "
"l        = [r, 0],  "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"from     = [s, none], "
"type     = [s, none, none], "
"bbshape  = [i, 1],  "
"sigx     = [r, 0],  "
"sigy     = [r, 0],  "
"xma      = [r, 0],  "
"yma      = [r, 0],  "
"width    = [r, 0],  "
"magnet   = [i, 0],  "
"charge   = [r, 1],  "
"angle    = [r, 0],  "
"copx     = [r, 0],  "
"copy     = [r, 0],  "
"alfxs    = [r, 0],  "
"alfys    = [r, 0],  "
"dxs      = [r, 0],  "
"dys      = [r, 0],  "
"dpxs     = [r, 0],  "
"dpys     = [r, 0],  "
"sigts    = [r, 1],  "
"siges    = [r, 1],  "
"cot      = [r, 0],  "
"copt     = [r, 0],  "
"slice    = [i, 1],  "
"iopt     = [i, 1],  "
"apertype = [s, circle, circle], "
"aperture = [r, {0}],  "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"cox      = [r, 0],  "
"coy      = [r, 0], "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"bbdir = [i, -1], "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true]; "
" "
"instrument: element none 0 24 "
"at       = [r, 1.e20], "
"l        = [r, 0], "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"tilt     = [r, 0],  "
"apertype = [s, circle, circle], "
"aperture = [r, {0}],  "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"magnet   = [i, 0],  "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"from     = [s, none], "
"type     = [s, none, none], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true]; "
" "
"marker: element none 0 25 "
"at       = [r, 1.e20], "
"l        = [r, 0], "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"type     = [s, none, none], "
"apertype = [s, circle, circle], "
"aperture = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"magnet   = [i, 0],  "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"from     = [s, none], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true]; "
" "
"gbend: element none 0 26 "
"at       = [r, 1.e20], "
"l        = [r, 0], "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"tilt     = [r, 0],  "
"angle    = [r, 0], "
"k0       = [r, 0],  "
"k0s      = [r, 0],  "
"k1       = [r, 0], "
"k1s      = [r, 0], "
"k2       = [r, 0], "
"k2s      = [r, 0], "
"e1       = [r, 0], "
"e2       = [r, 0], "
"ks       = [r, 0], "
"h1       = [r, 0], "
"h2       = [r, 0], "
"hgap     = [r, 0], "
"fint     = [r, 0, 0.5], "
"fintx    = [r, -1.0], "
"k3       = [r, 0], "
"k3s      = [r, 0], "
"fcsr     = [r, 0], "
"gmin     = [r, 0], "
"gmax     = [r, 0], "
"magnet   = [i, 1],  "
"slice    = [i, 1],  "
"apertype = [s, circle, circle], "
"aperture = [r, {0}], "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"from     = [s, none], "
"type     = [s, none, none], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, true, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true]; "
" "
"twcavity: element none 0 27 "
"at       = [r, 1.e20], "
"l        = [r, 0],  "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"volt     = [r, 0],  "
"lag      = [r, 0],  "
"freq     = [r, 0],  "
"harmon   = [i, 0],  "
"betrf    = [r, 0],  "
"pg       = [r, 0],  "
"shunt    = [r, 0],  "
"tfill    = [r, 0],  "
"eloss    = [r, 0],  "
"volterr  = [r, 0],  "
"lagerr   = [r, 0],  "
"nbin     = [i, 0],  "
"binmax   = [r, 0],  "
"magnet   = [i, 0],  "
"slice    = [i, 1],  "
"lfile    = [s, none, lfile],  "
"tfile    = [s, none, tfile], "
"apertype = [s, circle, circle], "
"aperture = [r, {0}],  "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"from     = [s, none], "
"type     = [s, none, none], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true], "
"psi             = [r, 0], "
"delta_lag       = [r, 0]; "
" "
"wire: element none 0 29 "
"at       = [r, 1.e20], "
"l        = [r, 0], "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"apertype = [s, circle, circle], "
"aperture = [r, {0}],  "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"magnet   = [i, 0],  "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"from     = [s, none], "
"type     = [s, none, none], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true]; "
" "
"slmonitor: element none 0 30 "
"at       = [r, 1.e20], "
"l        = [r, 0], "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"magnet   = [i, 0],  "
"apertype = [s, circle, circle], "
"aperture = [r, {0}],  "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"from     = [s, none], "
"type     = [s, none, none], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true]; "
" "
"blmonitor: element none 0 31 "
"at       = [r, 1.e20], "
"l        = [r, 0], "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"magnet   = [i, 0],  "
"apertype = [s, circle, circle], "
"aperture = [r, {0}],  "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"from     = [s, none], "
"type     = [s, none, none], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true]; "
" "
"imonitor: element none 0 32 "
"at       = [r, 1.e20], "
"l        = [r, 0], "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"magnet   = [i, 0],  "
"apertype = [s, circle, circle], "
"aperture = [r, {0}],  "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"from     = [s, none], "
"type     = [s, none, none], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true]; "
" "
"dipedge: element none 0 33 "
"at       = [r, 1.e20], "
"l        = [r, 0],  "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"tilt     = [r, 0],  "
"e1       = [r, 0],  "
"h        = [r, 0],  "
"hgap     = [r, 0],  "
"fint     = [r, 0, 0.5],  "
"slice    = [i, 1],  "
"apertype = [s, circle, circle], "
"aperture = [r, {0}], "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"magnet   = [i, 1],  "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"from     = [s, none], "
"type     = [s, none, none], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, true, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true]; "
" "
"changeref: element none 0 35 "
"at       = [r, 1.e20], "
"l        = [r, 0], "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"magnet   = [i, 0],  "
"patch_ang   = [r, {0, 0, 0}], "
"patch_trans = [r, {0, 0, 0}], "
"apertype = [s, circle, circle], "
"aperture = [r, {0}],  "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"from     = [s, none], "
"type     = [s, none, none], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true]; "
" "
"translation: element none 0 36 "
"at       = [r, 1.e20], "
"l        = [r, 0], "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"magnet   = [i, 0], "
"x     = [r, 0], y    = [r, 0], t    = [r, 0], "
"px    = [r, 0], py   = [r, 0], pt   = [r, 0], "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"from     = [s, none], "
"type     = [s, none, none], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true]; "
" "
"crabcavity: element none 0 37 "
"at       = [r, 1.e20], "
"l        = [r, 0],  "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"volt     = [r, 0],  "
"lag      = [r, 0],  "
"lagf     = [r, 0],  "
"freq     = [r, 0],  "
"harmon   = [i, 0],  "
"rv1      = [i, 0],  "
"rv2      = [i, 0],  "
"rv3      = [i, 0],  "
"rv4      = [i, 0],  "
"rph1     = [i, 0],  "
"rph2     = [i, 0],  "
"betrf    = [r, 0],  "
"pg       = [r, 0],  "
"shunt    = [r, 0],  "
"tfill    = [r, 0],  "
"eloss    = [r, 0],  "
"volterr  = [r, 0],  "
"lagerr   = [r, 0],  "
"nbin     = [i, 0],  "
"binmax   = [r, 0],  "
"magnet   = [i, 0],  "
"slice    = [i, 1],  "
"lfile    = [s, none, lfile],  "
"tfile    = [s, none, tfile], "
"apertype = [s, circle, circle], "
"aperture = [r, {0}],  "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"from     = [s, none], "
"type     = [s, none, none], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true], "
"n_bessel        = [i, 0], "
"tilt     = [r, 0];  "
" " 
"placeholder: element none 0 38 "
"at       = [r, 1.e20], "
"l        = [r, 0], "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"tilt     = [r, 0],  "
"lrad     = [r, 0],  "
"apertype = [s, circle, circle], "
"aperture = [r, {0}],  "
"aper_tol = [r, {0, 0, 0}], "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"magnet   = [i, 0],  "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"from     = [s, none], "
"type     = [s, none, none], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true]; "
" "
"tkicker: element none 0 39 "
"at       = [r, 1.e20], "
"l        = [r, 0],  "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"tilt     = [r, 0],  "
"lrad     = [r, 0],  "
"hkick    = [r, 0],  "
"vkick    = [r, 0],  "
"chkick   = [r, 0],  "
"cvkick   = [r, 0],  "
"chflag   = [i, 1],  "
"cvflag   = [i, 1],  "
"slice    = [i, 1],  "
"apertype = [s, circle, circle], "
"aperture = [r, {0}],  "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"magnet   = [i, 1],  "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"from     = [s, none], "
"type     = [s, none, none], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true]; "
" "
"hacdipole: element none 0 40 "
"at       = [r, 1.e20], "
"l        = [r, 0],  "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"from     = [s, none], "
"type     = [s, none, none], "
"volt     = [r, 0],  "
"lag      = [r, 0],  "
"freq     = [r, 0],  "
"harmon   = [i, 0],  "
"ramp1    = [i, 0],  "
"ramp2    = [i, 0],  "
"ramp3    = [i, 0],  "
"ramp4    = [i, 0],  "
"betrf    = [r, 0],  "
"pg       = [r, 0],  "
"shunt    = [r, 0],  "
"tfill    = [r, 0],  "
"eloss    = [r, 0],  "
"volterr  = [r, 0],  "
"lagerr   = [r, 0],  "
"nbin     = [i, 0],  "
"binmax   = [r, 0],  "
"magnet   = [i, 0],  "
"slice    = [i, 1],  "
"lfile    = [s, none, lfile],  "
"tfile    = [s, none, tfile], "
"apertype = [s, circle, circle], "
"aperture = [r, {0}],  "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true], "
"n_bessel        = [i, 0]; "
" "
"vacdipole: element none 0 41 "
"at       = [r, 1.e20], "
"l        = [r, 0],  "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"from     = [s, none], "
"type     = [s, none, none], "
"volt     = [r, 0],  "
"lag      = [r, 0],  "
"freq     = [r, 0],  "
"harmon   = [i, 0],  "
"ramp1    = [i, 0],  "
"ramp2    = [i, 0],  "
"ramp3    = [i, 0],  "
"ramp4    = [i, 0],  "
"betrf    = [r, 0],  "
"pg       = [r, 0],  "
"shunt    = [r, 0],  "
"tfill    = [r, 0],  "
"eloss    = [r, 0],  "
"volterr  = [r, 0],  "
"lagerr   = [r, 0],  "
"nbin     = [i, 0],  "
"binmax   = [r, 0],  "
"magnet   = [i, 0],  "
"slice    = [i, 1],  "
"lfile    = [s, none, lfile],  "
"tfile    = [s, none, tfile], "
"apertype = [s, circle, circle], "
"aperture = [r, {0}],  "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true], "
"n_bessel        = [i, 0]; "
" "
"nllens: element none 0 42 "
"at       = [r, 1.e20], "
"l        = [r, 0], "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"knll     = [r, 0], "
"cnll     = [r, 0], "
"tilt     = [r, 0],  "
"lrad     = [r, 0],  "
"apertype = [s, circle, circle], "
"aperture = [r, {0}],  "
"aper_tol = [r, {0, 0, 0}], "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"magnet   = [i, 0],  "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"from     = [s, none], "
"type     = [s, none, none], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true]; "
" "
"rfmultipole: element none 0 43 "
"at       = [r, 1.e20], "
"l        = [r, 0], "
"kmax     = [r, 0], "
"kmin     = [r, 0], "
"calib    = [r, 0], "
"polarity = [r, 0], "
"noise     = [r, 0], "
"noisemax    = [i, 1],  "
"npeak     = [r, {0}], "
"ntune     = [r, {0}], "
"nlag     = [r, {0}], "
"tilt     = [r, 0],  "
"lrad     = [r, 0], "
"knl      = [r, {0}], "
"ksl      = [r, {0}], "
"apertype = [s, circle, circle], "
"aperture = [r, {0}],  "
"aper_offset = [r, {0}], "
"aper_tol = [r, {0, 0, 0}], "
"slice    = [i, 1],  "
"slot_id  = [i, none], "
"assembly_id = [i, none], "
"mech_sep = [r, 0], "
"v_pos = [r, 0], "
"magnet   = [i, 1],  "
"model           = [i, -1], "
"method          = [i, -1], "
"exact           = [i, -1, 1], "
"nst             = [i, -1], "
"from     = [s, none], "
"type     = [s, none, none], "
"permfringe      = [l, false, true], "
"bend_fringe     = [l, false, true], "
"kill_ent_fringe = [l, false, true], "
"kill_exi_fringe = [l, false, true], "
"freq       = [r, 0], "   // RF-Multipole
"volt       = [r, 0], "   // RF-Multipole
"lag        = [r, 0], "   // RF-Multipole
"harmon     = [i, 0], "   // RF-Multipole
"pnl        = [r, {0}], " // RF-Multipole
"psl        = [r, {0}]; " // RF-Multipole
" "  
"emit: emit none 0 0 "
"deltap   = [r, 0], "
"tol      = [r, 1.000001, 0]; "
" "
"ealign: error none 0 0 "
"dx       = [r, 0], "
"dy       = [r, 0], "
"ds       = [r, 0], "
"dphi     = [r, 0], "
"dtheta   = [r, 0], "
"dpsi     = [r, 0], "
"mrex     = [r, 0], "
"mrey     = [r, 0], "
"mredx    = [r, 0], "
"mredy    = [r, 0], "
"arex     = [r, 0], "
"arey     = [r, 0], "
"mscalx   = [r, 0], "
"mscaly   = [r, 0]; "
" "
"eoption: error none 0 0 "
"seed     = [i, 123456789], "
"add      = [l, true]; "
" "
"eprint: error none 0 0 "
"full     = [l, false]; "
" "
"esave: error none 0 0 "
"file     = [s, esave]; "
" "
"seterr: error none 0 0 "
"table = [s, none], "
"file = [s, errfil, errfil]; "
" "
"efcomp: error none 0 0 "
"order    = [i, -1], "
"hyster   = [i, 0], "
"radius   = [r, 0], "
"hcoeffn  = [r, {0}], "
"hcoeffs  = [r, {0}], "
"dkn      = [r, {0}], "
"dks      = [r, {0}], "
"dknr     = [r, {0}], "
"dksr     = [r, {0}], "
"rfm_freq   = [r, 0], "   /* AL: frequency of the RF-multipoles main mode */
"rfm_harmon = [i, 0], "   /* AL: harmonic number of the RF-multipoles main mode */
"rfm_lag    = [r, 0], "   /* AL: lag for the RF-multipoles main mode */
"dpn        = [r, {0}], " /* AL: abssolute phase of the normal components of the multipole */
"dps        = [r, {0}]; " /* AL: absolute phase of the skew components of the multipole */
" "
"sddsin: sdds none 0 0 "
"table    = [s, none, none], "
"file     = [s, none]; "
" "
"sddsout: sdds none 0 0 "
"table    = [s, none, none], "
"file     = [s, none]; "
" "
" "
"ibs: ibs none 0 0 "
"tolerance= [r, 1.e-7], "
"file     = [s, ibs, ibs], "
"steps    = [i, 50]; "
" "
"touschek: touschek none 0 0 "
"tolerance= [r, 1.e-7], "
"file     = [s, touschek, touschek]; "
" "
"makethin: makethin none 0 0 "
"style    = [s, teapot, teapot], "
"sequence = [s, none, none], "
"makeconsistent = [l, false, true], " /*hbu */
"minimizeparents = [l, true, false]; " /*hbu */
" "
"survey: survey none 0 0 "
"x0       = [r, 0],   y0     = [r, 0], z0       = [r, 0], "
"theta0   = [r, 0], phi0     = [r, 0], psi0     = [r, 0], "
"file     = [s, none, survey], "
"table    = [s, none, survey], "
"sequence = [s, none, sequence]; "
" "
"twiss: twiss none 0 0 "
"betx     = [r, 0], alfx     = [r, 0], mux      = [r, 0], "
"bety     = [r, 0], alfy     = [r, 0], muy      = [r, 0], "
"x        = [r, 0], px       = [r, 0],         "
"y        = [r, 0], py       = [r, 0],         "
"t        = [r, 0], pt       = [r, 0],         "
"dx       = [r, 0], dpx      = [r, 0],         "
"dy       = [r, 0], dpy      = [r, 0],         "
"wx       = [r, 0], phix     = [r, 0], dmux     = [r, 0], "
"wy       = [r, 0], phiy     = [r, 0], dmuy     = [r, 0], "
"ddx      = [r, 0], ddpx     = [r, 0], "
"ddy      = [r, 0], ddpy     = [r, 0], "
"r11      = [r, 0], r12      = [r, 0], "
"r21      = [r, 0], r22      = [r, 0], "
"energy   = [r, 0], "
"chrom    = [l, false, true], "
"file     = [s, none, twiss], "
"save     = [s, none, twiss], "
"table    = [s, none, twiss], "
"beta0    = [s, none, beta0], "
"re11     = [r, 1], re12     = [r, 0], re13     = [r, 0], "
"re14     = [r, 0], re15     = [r, 0], re16     = [r, 0], "
"re21     = [r, 0], re22     = [r, 1], re23     = [r, 0], "
"re24     = [r, 0], re25     = [r, 0], re26     = [r, 0], "
"re31     = [r, 0], re32     = [r, 0], re33     = [r, 1], "
"re34     = [r, 0], re35     = [r, 0], re36     = [r, 0], "
"re41     = [r, 0], re42     = [r, 0], re43     = [r, 0], "
"re44     = [r, 1], re45     = [r, 0], re46     = [r, 0], "
"re51     = [r, 0], re52     = [r, 0], re53     = [r, 0], "
"re54     = [r, 0], re55     = [r, 1], re56     = [r, 0], "
"re61     = [r, 0], re62     = [r, 0], re63     = [r, 0], "
"re64     = [r, 0], re65     = [r, 0], re66     = [r, 1], "
"centre   = [l, false, true], "
"ripken   = [l, false, true], "
"sectormap= [l, false, true], "
"sectortable= [s, none, sectortable], "
"sectorfile= [s, sectormap, sectormap], "
"rmatrix  = [l, false, true], "
"sequence = [s, none, sequence], "
"line     = [s, none, line], "
"range    = [s, #s/#e, none], "
"useorbit = [s, default, default], "
"keeporbit= [s, default, default], "
"tolerance= [r, 1.e-6], "
"deltap   = [s, none], "
"notable  = [l, false, true]; "
" "
"match: match match 1 0 "
"use_macro= [l, false, true], "
"use_ptcknob = [l, false, true], "
"betx     = [r, {0}], alfx     = [r, {0}], mux      = [r, {0}], "
"bety     = [r, {0}], alfy     = [r, {0}], muy      = [r, {0}], "
"x        = [r, {0}], px       = [r, {0}], "
"y        = [r, {0}], py       = [r, {0}], "
"t        = [r, {0}], pt       = [r, {0}], "
"dx       = [r, {0}], dpx      = [r, {0}], "
"dy       = [r, {0}], dpy      = [r, {0}], "
"wx       = [r, {0}], phix     = [r, {0}], dmux     = [r, {0}], "
"wy       = [r, {0}], phiy     = [r, {0}], dmuy     = [r, {0}], "
"ddx      = [r, {0}], ddpx     = [r, {0}], "
"ddy      = [r, {0}], ddpy     = [r, {0}], "
"r11      = [r, {0}], r12      = [r, {0}], "
"r21      = [r, {0}], r22      = [r, {0}], "
"energy   = [r, {0}], "
"mvar1    = [r, {0}], mvar2    = [r, {0}], "
"mvar3    = [r, {0}], mvar4    = [r, {0}], "
"sequence = [s, {none}], "
"beta0    = [s, {none}], "
"rmatrix  = [l, false, true], "
"chrom    = [l, false, true], "
"range    = [s, {none}], "
"deltap   = [r, {0}], "
"useorbit = [s, {default}, {default}], "
"keeporbit= [s, {default}, {default}], "
"vlength  = [l, false, true], "
"slow     = [l, false, true], "  /* makes match use the twiss table */
"orbit    = [l, false, true]; "
" "
"cell: match match 0 0; "
" "
"endmatch:  match match 2 0 "
"knobfile = [s, none, none], "
"increment = [s, none, increment]; "
" "
"migrad: match match 0 0 "
"tolerance= [r, 0.000001], "
"calls    = [i, 1000], "
"strategy = [i, 2]; "
" "
"simplex: match match 0 0 "
"tolerance= [r, 0.000001], "
"calls    = [i, 1000]; "
" "
"siman: match match 0 0 "
"tolerance= [r, 0.000001, 0.000001], "
"calls    = [i, 1000, 1000]; "
" "
"constraint: match match 0 0 "
"expr     = [s, none, none], "
"name    = [s, none, none], "
"weight   = [r, 1.0], "
"range    = [s, #s/#e, none], "
"class    = [s, none, none], "
"pattern  = [s, any, none], "
"betx     = [c, 0], alfx     = [c, 0], mux      = [c, 0], "
"bety     = [c, 0], alfy     = [c, 0], muy      = [c, 0], "
"x        = [c, 0], px       = [c, 0], "
"y        = [c, 0], py       = [c, 0], "
"t        = [c, 0], pt       = [c, 0], "
"dx       = [c, 0], dpx      = [c, 0], "
"dy       = [c, 0], dpy      = [c, 0], "
"wx       = [c, 0], phix     = [c, 0], dmux     = [c, 0], "
"wy       = [c, 0], phiy     = [c, 0], dmuy     = [c, 0], "
"ddx      = [c, 0], ddpx     = [c, 0], "
"ddy      = [c, 0], ddpy     = [c, 0], "
"r11      = [c, 0], r12      = [c, 0], "
"r21      = [c, 0], r22      = [c, 0], "
"energy   = [c, 0], "
"mvar1    = [c, 0], mvar2    = [c, 0], "
"mvar3    = [c, 0], mvar4    = [c, 0], "
"re11     = [c, 1], re12     = [c, 0], re13     = [c, 0], "
"re14     = [c, 0], re15     = [c, 0], re16     = [c, 0], "
"re21     = [c, 0], re22     = [c, 1], re23     = [c, 0], "
"re24     = [c, 0], re25     = [c, 0], re26     = [c, 0], "
"re31     = [c, 0], re32     = [c, 0], re33     = [c, 1], "
"re34     = [c, 0], re35     = [c, 0], re36     = [c, 0], "
"re41     = [c, 0], re42     = [c, 0], re43     = [c, 0], "
"re44     = [c, 1], re45     = [c, 0], re46     = [c, 0], "
"re51     = [c, 0], re52     = [c, 0], re53     = [c, 0], "
"re54     = [c, 0], re55     = [c, 1], re56     = [c, 0], "
"re61     = [c, 0], re62     = [c, 0], re63     = [c, 0], "
"re64     = [c, 0], re65     = [c, 0], re66     = [c, 1], "
"alfa     = [c, 0],"
"beta11     = [c, 0], alfa11    = [c, 0], mu1      = [c, 0], "
"beta22     = [c, 0], alfa22    = [c, 0], mu2      = [c, 0], "
"beta33     = [c, 0], alfa33    = [c, 0], mu3      = [c, 0], "
"disp1      = [c, 0], disp2     = [c, 0], disp3    = [c, 0], , disp4    = [c, 0],"
"beta0    = [s, none], "
"sequence = [s, none]; "
" "
"couple: match match 0 0 "
"range    = [s, none, none], "
"mux      = [r, 0], muy      = [r, 0]; "
" "
"fix: match match 0 0 "
"name     = [s, none, none]; "
" "
"level: match match 0 0 "
"level    = [i, 2]; "
" "
"vary: match match 0 0 "
"name     = [s, none, none], "
"step     = [r, 0.0], "
"slope    = [i, 0], "
"init     = [r, 0.0], "
"knob     = [s, knob, knob], "
"lower    = [r, -1.e20], "
"opt      = [r,  0    ], "
"upper    = [r,  1.e20]; "
" "
"weight: match match 0 0 "
"betx     = [r, 1.0], "
"alfx     = [r, 10.0], "
"mux      = [r, 10.0], "
"bety     = [r, 1.0], "
"alfy     = [r, 10.0], "
"muy      = [r, 10.0], "
"x        = [r, 10.0], "
"px       = [r, 100.0], "
"y        = [r, 10.0], "
"py       = [r, 100.0], "
"t        = [r, 10.0], "
"pt       = [r, 100.0], "
"dx       = [r, 10.0], "
"dpx      = [r, 100.0], "
"dy       = [r, 10.0], "
"dpy      = [r, 100.0], "
"wx       = [r, 1.0], "
"phix     = [r, 1.0], "
"dmux     = [r, 1.0], "
"wy       = [r, 1.0], "
"phiy     = [r, 1.0], "
"dmuy     = [r, 1.0], "
"ddx      = [r, 1.0], "
"ddpx     = [r, 1.0], "
"ddy      = [r, 1.0], "
"ddpy     = [r, 1.0], "
"mvar1    = [r, 10.0], "
"mvar2    = [r, 10.0], "
"mvar3    = [r, 10.0], "
"mvar4    = [r, 10.0], "
"r11      = [r, 1.0], "
"r12      = [r, 1.0], "
"r21      = [r, 1.0], "
"r22      = [r, 1.0], "
"energy   = [r, 1.0], "
"circ     = [r, 1.0], "
"i1       = [r, 1.0], "
"i2       = [r, 1.0], "
"i3       = [r, 1.0], "
"i4       = [r, 1.0], "
"i5       = [r, 1.0], "
"i5i2     = [r, 1.0], "
"i5i1     = [r, 1.0], "
"re11     = [r, 1.0], re12     = [r, 1.0], re13     = [r, 1.0], "
"re14     = [r, 1.0], re15     = [r, 1.0], re16     = [r, 1.0], "
"re21     = [r, 1.0], re22     = [r, 1.0], re23     = [r, 1.0], "
"re24     = [r, 1.0], re25     = [r, 1.0], re26     = [r, 1.0], "
"re31     = [r, 1.0], re32     = [r, 1.0], re33     = [r, 1.0], "
"re34     = [r, 1.0], re35     = [r, 1.0], re36     = [r, 1.0], "
"re41     = [r, 1.0], re42     = [r, 1.0], re43     = [r, 1.0], "
"re44     = [r, 1.0], re45     = [r, 1.0], re46     = [r, 1.0], "
"re51     = [r, 1.0], re52     = [r, 1.0], re53     = [r, 1.0], "
"re54     = [r, 1.0], re55     = [r, 1.0], re56     = [r, 1.0], "
"re61     = [r, 1.0], re62     = [r, 1.0], re63     = [r, 1.0], "
"re64     = [r, 1.0], re65     = [r, 1.0], re66     = [r, 1.0], "
"alfa     = [r, 1.0],"
"dumm     = [r, 1.0]; "
" "
"lmdif: match match 0 0 "
"tolerance= [r, 0.000001, 0.000001], "
"calls    = [i, 1000, 1000]; "
" "
"jacobian: match match 0 0 "
"tolerance= [r, 0.000001, 0.000001], "
"calls    = [i, 30, 30], "
"cool     = [r, 0.0], "
"balance  = [r, 0.5], "
"random   = [r, 0.0], "
"repeat   = [r, 1], "
"bisec   = [i, 1], "
"cond    = [r, 1E-8], "
"knobfile = [s, none, none], "
"jacfile  = [s, none, none], "
"strategy = [i, 3]; "
" "
"rmatrix: match match 0 0 "
"range    = [s, #s/#e, none], "
"rm       = [r, {0}], "
"weight   = [r, {0}]; "
" "
"tmatrix: match match 0 0 "
"range    = [s, #s/#e, none], "
"tm       = [r, {0}], "
"weight   = [r, {0}]; "
" "
"global: match match 0 0 "
"q1       = [c, 0],   q2        = [c, 0], "
"dq1      = [c, 0],   dq2       = [c, 0], "
"ddq1     = [c, 0],   ddq2      = [c, 0], "
"dq1de1   = [c, 0],   dq1de2    = [c, 0], "
"dq2de2   = [c, 0],   gammatr   = [c, 0], "
"alfa     = [c, 0],"
"sequence = [s, none]; "
" "
"gweight: match match 0 0 "
"q1       = [r, 10],  q2        = [r, 10], "
"dq1      = [r, 1],   dq2       = [r, 1], "
"ddq1     = [r, 0.1], ddq2      = [r, 0.1], "
"dq1de1   = [r, 0.1], dq1de2    = [r, 0.1], "
"dq2de2   = [r, 1],   alfa      = [r, 1.0],"
"gammatr   = [r, 1]; "
" "
"sequence: sequence none 0 0 "
"at       = [r, 1.e20], "
"from     = [s, none], "
"refpos   = [s, none], "
"l        = [r, 0], "
"add_pass = [r, 0], "
"next_sequ= [s, none], "
"refer    = [s, centre, centre]; "
" "
"endsequence: sequence none 0 0; "
" "
"sxfread: sxf none 0 0 "
"file = [s, sxf_in, sxf_in]; "
" "
"sxfwrite: sxf none 0 0 "
"file = [s, sxf_out, sxf_out]; "
" "
"track: track track 1 0 "
"deltap   = [r, 0], "
"onepass  = [l, false, true], "
"update   = [l, false, true], "
"damp     = [l, false, true], "
"quantum  = [l, false, true], "
"dump     = [l, false, true], "
/* "fast     = [l, false, true], " */
"aperture = [l, false, true], "
"onetable = [l, false, true], "
"recloss  = [l, false, true], "
"file     = [s, track, track], "
"extension= [s, none, none]; "
" "
"dynap: track track 0 0 "
"turns   = [i, 64], "
"fastune  = [l, false,true], "
"lyapunov = [r, 1.e-7], "
"maxaper= [r, {0.1, 0.01, 0.1, 0.01, 1., 0.1}], "
"damp     = [l, true], "
"quantum  = [l, true], "
"orbit    = [l, true]; "
" "
"endtrack: track track 2 0; "
" "
"run: track track 0 0 "
"maxaper= [r, {0.1, 0.01, 0.1, 0.01, 1., 0.1}], "
"turns    = [i, 1], "
"ffile    = [i, 1]; "
" "
"start: track track 0 0 "
"x        = [r, 0], "
"px       = [r, 0], "
"y        = [r, 0], "
"py       = [r, 0], "
"t        = [r, 0], "
"pt       = [r, 0], "
"fx       = [r, 0], "
"phix     = [r, 0], "
"fy       = [r, 0], "
"phiy     = [r, 0], "
"ft       = [r, 0], "
"phit     = [r, 0]; "
" "
"ripple: track track 0 0 "
"variable = [s, none], "
"amplitude= [r, {0}], "
"frequency= [r, {0}], "
"phase    = [r, {0}]; "
" "
"observe: track track 0 0 "
"place    = [s, none]; "
" "
"use_macro: match match 0 0 "
"name     = [s, none, none]; "
" "
"setvars_lin: control none 0 0 "
"table = [s, none], "
"row1 = [i, 0], "
"row2 = [i, 0], "
"param = [s, t]; "
" "
;
