#include "MadxPlotter.h"
/*Piotr Skowronski, CERN*/

#include <TApplication.h>
#include <TSystem.h>

#include <TString.h>
#include <TError.h>
#include <TH2.h>
#include <TNtuple.h>
#include <TObjArray.h>
#include <TFile.h>
#include <Riostream.h>


ClassImp(MadxPlotter)

MadxPlotter* MadxPlotter::fgMadxPlotter = 0x0;
Int_t        MadxPlotter::fgkDebug      = 0;

//______________________________________________________________________________________

MadxPlotter::MadxPlotter():
 fPlotNo(0),
 fPlots(new TObjArray()),
 fTrees(new TObjArray()),
 fFile(0x0)
{
  //ctor
  
  if (!gApplication)
      TApplication::CreateApplication();
  
//  fFile = TFile::Open("rplot.root","read");
  if (fFile)
   {  
     Fatal("MadxPlotter","File rplot.root already exist");
   }
  
  fFile = TFile::Open("rplot.root","recreate");

}
//______________________________________________________________________________________

MadxPlotter::~MadxPlotter()
{
//dtor
  FinishAll();
  fPlots->SetOwner();
  fTrees->SetOwner();
  
  delete fPlots;
  delete fTrees;
  delete fFile;
}


//______________________________________________________________________________________

MadxPlotter* MadxPlotter::Instance()
{
  //returns plotter object
  if (fgMadxPlotter == 0x0) fgMadxPlotter = new MadxPlotter();
  
  
  return fgMadxPlotter;
}

//______________________________________________________________________________________

void MadxPlotter::Delete()
{
  delete fgMadxPlotter;
  fgMadxPlotter = 0x0;
}
//______________________________________________________________________________________

Int_t  MadxPlotter::NewPlot()
{
  //creates new ntuple and monitoring histograms
  
  if (fTrees->At(fPlotNo)) 
   {
     Finish();
   }

  fPlotNo++;
  TString plotname("Tracking");
  plotname+=fPlotNo;

  Info("NewPlot","Crating new rplot %d named %s",fPlotNo,plotname.Data());

  AddPlot(fPlotNo,6,plotname);
  return fPlotNo;
}

//______________________________________________________________________________________

void  MadxPlotter::Finish()
{
//finishes a plotting session
  

  Info("Finish","Finishing %d",fPlotNo);
  ((TNtuple*)fTrees->At(fPlotNo))->Write(0,TObject::kOverwrite);
  delete fTrees->RemoveAt(fPlotNo);
  
  

}
//______________________________________________________________________________________

void  MadxPlotter::FinishAll()
{
//dinishes a plotting session
 
 if (fFile) 
  {
    fFile->cd();
  } 
 else
  {
    Info("FinishAll","File is null");
    return;
  } 
 
 TIter nexth(fPlots);
 TH1* h = 0x0;
 while (( h = (TH1*)nexth() ))
  {
    h->Write(0,TObject::kOverwrite);
  } 

 TIter nextnt(fTrees);
 TNtuple* ntuple = 0x0;
 while (( ntuple = (TNtuple*)nextnt() ))
   {
     ntuple->Write(0,TObject::kOverwrite);
   }   
   
  
  
}
//______________________________________________________________________________________


Int_t MadxPlotter::AddPlot(Int_t handler, Int_t nparams, TString name)
{
//Adds another plot
  Info("AddPlot","Adding %d",handler);
  if (fPlots->At(handler) != 0x0) 
   {
     Info("AddPlot","Handle %d already existing. Deleting it.",handler);
     delete fPlots->RemoveAt(handler);
   }

  
//  TH2D* h = new TH2D(name,name,100,-0.001,0.001,100,-0.001,0.001);
//  h->SetDirectory(fFile);
//  fPlots->AddAtAndExpand(h,handler);
  
  fFile->cd();
  TNtuple* ntpl = new TNtuple(name,name,"trackno:obspoint:turn:x:xp:y:yp:dPoverP:p:l");
//  Info("AddPlot","%#x",ntpl);
  fTrees->AddAtAndExpand(ntpl,handler);
  
 return 0;
}

Int_t MadxPlotter::Fill(Int_t obsp, Double_t & betax, Double_t & alfax, 
                                    Double_t & betay, Double_t & alfay,
                                    Double_t & betaz, Double_t & alfaz, Double_t & length)
{
  // fills a twiss
 if (GetDebug() > 9) 
  {
   Info("Fill","Got TWISS Data at observation point %d:",obsp);
   Info("Fill","    %12.8f %12.8f %12.8f %12.8f %12.8f %12.8f %12.8f",
                    betax, alfax, betay, alfay, betaz, alfaz, length);
  }
 
 return 0; 
}


Int_t MadxPlotter::Fill(Int_t particleno, Int_t obsp, Int_t turn, Double_t x, Double_t xp, 
                        Double_t y, Double_t yp, Double_t dpOverP, Double_t p, Double_t l)
{
 //adds one entry to a plot (tree / histo)

 if (GetDebug() > 9) 
  {
   Info("Fill","fPlotNo = %d",fPlotNo);
   Info("Fill","Got Data for particle %d at observation point %d turn %d:",particleno, obsp, turn);
   Info("Fill","    %12.8f %12.8f %12.8f %12.8f %12.8f   %12.8f %12.8f",
                    x,     xp,    y,     yp,    dpOverP, p,     l);
  }
 if (fTrees->At(fPlotNo) == 0x0) 
   { 
     if (GetDebug() > 2) 
      {
        Info("Fill","There is not tree in array at %d. Adding it.",fPlotNo);
        
      }
     TString plotname("Tracking");
     plotname+=fPlotNo;
     AddPlot(fPlotNo,6,plotname);
   }  
              
// ((TH1*)fPlots->At(handler))->Fill(x,y); only if added the online monitoring 
 
 
 TNtuple* nt = ((TNtuple*)fTrees->At(fPlotNo));
//  Info("Fill","%#x",nt);

 nt->Fill(particleno,obsp, turn, x,xp, y,yp, dpOverP, p, l);
 return 0;
}

/*************************************************************************************/ 

void MadxPlotter::PressAnyKey()
{ 
 //small utility function that helps to make comfortable macros
  char c;
  int nread = -1;
  fcntl(0,  F_SETFL, O_NONBLOCK);
  ::Info("","Press Any Key to continue ...");
  while (nread<1) 
   {
     nread = read(0, &c, 1);
     gSystem->ProcessEvents();
   }
}


/*************************************************************************************/ 
/*************************************************************************************/ 
/*************************************************************************************/ 
/*************************************************************************************/ 


extern "C" void madxplotter_newrplot()
{
/*adds new plotter*/
  MadxPlotter::Instance()->NewPlot();
}

extern "C" void madxplotter_rplotfinish()
{
/*terminates plotter*/
  MadxPlotter::Instance()->Finish();/*writes and deletes the current plot*/
  printf("rplot: WARNING Finish deletes the instance as temporary bug fix\n");
  MadxPlotter::Instance()->Delete();/*writes and deletes the current plot*/
}

extern "C" void madxplotter_plottrack(Int_t particleno, Int_t obsp, Int_t turn, Double_t x, Double_t xp, //this fills an track entry 
                                      Double_t y, Double_t yp, Double_t dpOverP, Double_t p, Double_t l)
{
  MadxPlotter::Instance()->Fill(particleno, obsp,turn, x,  xp,  y,  yp,  dpOverP, p ,l);
}
