#ifndef MADXPLOTTER_H
#define MADXPLOTTER_H
/*Piotr Skowronski, CERN*/

#include <TObject.h>

class TH2D;
class TTree;
class TString;
class TObjArray;
class TFile;

class MadxPlotter: public TObject
{
//class that handles tracking data storage and monitoring
  
 private: 
   MadxPlotter();
   virtual ~MadxPlotter();
   
 public:  
     
   static MadxPlotter* Instance();
   static void Delete();
   
   Int_t  Plotter(){return 0;}
   Int_t  NewPlot();   
   void   FinishAll();

   void   Finish();
   
   Int_t AddPlot(Int_t handler, Int_t nparams, TString name);
   Int_t Fill(Int_t particleno, Int_t obsp, Int_t turn, Double_t x, Double_t xp, //this fills an track entry 
              Double_t y, Double_t yp, Double_t dpOverP, Double_t p, Double_t l);

   Int_t Fill(Int_t obsp, Double_t & betay, Double_t & alfay, 
                          Double_t & betay, Double_t & alfay,
                          Double_t & betaz, Double_t & alfaz, Double_t & length);

   
   static void PressAnyKey();//small utility function that helps to make comfortable macros
   
   static Int_t GetDebug(){return fgkDebug;}
   static void SetDebug(Int_t deblevel){fgkDebug = deblevel;}
 protected:
   
   static MadxPlotter * fgMadxPlotter;// singleton MAD-X Plotter
   static Int_t         fgkDebug;//debug flag
   
   Int_t       fPlotNo;//number of the active plot
   TObjArray*  fPlots;//array with plots
   TObjArray*  fTrees;//array with trees
   TFile*      fFile;
   
  ClassDef(MadxPlotter,0)
};

extern "C" void madxplotter_newrplot();/*adds new plotter*/
extern "C" void madxplotter_rplotfinish(); /*writes and deletes the current plot*/
extern "C" void madxplotter_plottrack(Int_t particleno, Int_t obsp, Int_t turn, Double_t x, Double_t xp, //this fills an track entry 
                                      Double_t y, Double_t yp, Double_t dpOverP, Double_t p, Double_t l);

#endif
