#include "MadxViewer.h"
/*Piotr Skowronski, CERN*/

#include "MadxViewerPar.h"
#include "MadxViewerParAtEl.h"

#include <TApplication.h>
#include <TRootEmbeddedCanvas.h>
#include <TGLabel.h>
#include <TGNumberEntry.h>
#include <TGTab.h>
#include <TGedEditor.h>
#include <TCanvas.h>
#include <TError.h>
#include <TStyle.h>
#include <TGTableLayout.h>
#include <TGFileDialog.h>
#include <TROOT.h>
#include <TObjArray.h>

#include <dlfcn.h>
#include <float.h>

ClassImp(MadxViewer)


MadxViewer* MadxViewer::fgMadxViewer = 0x0;


const char* MadxViewer::fgkSaveAsTypes[]  =  
                    { 
	  "PostScript",   "*.ps",
                      "Encapsulated PostScript", "*.eps",
                      "PDF",          "*.pdf",
                      "SVG",          "*.svg",
                      "GIF",          "*.gif",
                      "ROOT macros",  "*.C",
                      "ROOT files",   "*.root",
                      "XML",          "*.xml",
                      "PNG",          "*.png",
                      "XPM",          "*.xpm",
                      "JPEG",         "*.jpg",
                      "TIFF",         "*.tiff",
                      "XCF",          "*.xcf",
                      "All files",    "*",
                      0,              0 
	};



//______________________________________________________________________________________


MadxViewer* MadxViewer::Instance()
{
  //returns Viewer object
  
  if (fgMadxViewer)  return fgMadxViewer;
  
  //Otherwise we start all the machine
  
  if (!gApplication)
   {
      ::Info("MadxViewer::Instance","Creating application");
      char *a = StrDup("RMAD");
      char *argv[1];
      Int_t argc = 1;
      argv[0] = a;
      new TApplication("RootApp", &argc, argv, 0, 0);
      delete [] a;
   }  

  if (gROOT->IsBatch())
   {
     ::Error("MadxViewer::Instance","We are in batch mode, can not create windows.");
     return 0x0;
   }


  if (gClient == 0x0)
   {
     ::Error("MadxViewer::Instance","gClient is NULL");
     return 0x0;
   }
  
  gStyle->SetCanvasColor(0);
  gStyle->SetPalette(1,0);
  gStyle->SetOptTitle(kFALSE);
   
  fgMadxViewer = new MadxViewer();
  
  Bool_t initok = fgMadxViewer->Init();
  
  if (initok)
   {
     ::Info("MadxViewer::Instance","Running RMAD");
     gApplication->Run();
   }
  else
   {
     ::Error("MadxViewer::Instance","Error occured during application initialization.");
   } 
  delete fgMadxViewer; fgMadxViewer = 0x0;//should be NULL here 
  
  delete gApplication;
  gApplication = 0x0;
  
  ::Info("MadxViewer::Instance","RMAD Run Finished");
  
  return fgMadxViewer;
}

//______________________________________________________________________________________


MadxViewer::MadxViewer():
 fMainFrame(0x0),
 fTab(0x0),
 fTabContainer(0x0),
 fParAtCanvas(0x0),
 fViewerPar(new MadxViewerPar(this)),
 fViewerParAtEl(new MadxViewerParAtEl(this)),
 fParSetter(new MadxViewerParSetter(this)),
 fGEditor(0x0),
 fActiveAtEl(kTRUE)
{
//ctor
//  Info("XXX","%#x",fViewerPar);
}
//______________________________________________________________________________________

Bool_t MadxViewer::Init()
{

   void *handle;
   char *error;
   
   /*simple check if the library was linked dynamically*/
   handle = dlopen( 0,  RTLD_GLOBAL | RTLD_LAZY);
//   handle = dlopen( 0 ,RTLD_NOW );
/*   printf("0 loading Done\n");*/
   if (handle == 0x0) 
    {
      error = dlerror();
      if (error) Error("Init", "%s\n", error);
      return kFALSE;
    }
   
  void * fctn = 0x0;
  
  fctn = dlsym(handle, "w_ptc_getnelements_");


  Info("Init","Dupa Got w_ptc_getnelements_ function at address %#x", fctn);
  
  if ((error = dlerror()) != NULL)  
   {
      Error("Init", "%s\n", error);
      return kFALSE;
   }
   
  Info("Init","Got w_ptc_getnelements_ function %#x", fctn);
  ///////////////////////////////////////////////////////////////
   
  if (fMainFrame == 0x0) Create();
  
  Info("Init","Checking for no. of elements in the line");
  Int_t nelem = w_ptc_getnelements(); 
  Info("Init","There is %d elements in the line",nelem);
  
  if (nelem <= 0)
   {
     Error("Init","PTC reports no elements");
     return kFALSE;
   }

  ReadParametrs();

  
  fViewerPar->Init();
  fViewerParAtEl->Init();


  ConnectSignals2Slots();

  fMainFrame->MapSubwindows();
  fMainFrame->Resize(fMainFrame->GetDefaultSize());
  fMainFrame->MapWindow();
  fMainFrame->MoveResize(10,10,900,600);
   
  fViewerParAtEl->AllowDrawSame();//too hide 2D options since by default 1D is plotted
  
  return kTRUE;

}
//______________________________________________________________________________________

MadxViewer::~MadxViewer()
{
//dtor
  Info("~MadxViewer","Deleting");

  delete fViewerPar;
  delete fViewerParAtEl;
  delete fParSetter;

  Info("~MadxViewer"," 2 ");
  delete fGEditor;

  Info("~MadxViewer"," 3 ");
 
//  delete fTab;
  if (fMainFrame) fMainFrame->Cleanup();//takes care of all the suframes 
  delete fMainFrame; 

  Info("~MadxViewer"," END ");
  
}
//______________________________________________________________________________________

void MadxViewer::Create()
{

  if (!gApplication)
   {
      ::Info("test","Creating application");
      char *a = StrDup("RMAD");
      char *argv[1];
      Int_t argc = 1;
      argv[0] = a;
      new TApplication("RootApp", &argc, argv, 0, 0);
      delete [] a;
   }  

  if (gROOT->IsBatch())
   {
     ::Error("test","We are in batch mode, can not create windows.");
     return;
   }

  if (gClient == 0x0)
   {
     gClient = new TGClient();
   }

   
   if (gClient == 0x0)
    {
      ::Error("test","gClient is NULL");
      return;
    }

   ULong_t ucolor;        // will reflect user color changes
   gClient->GetColorByName("#ffffff",ucolor);


   fMainFrame = new TGMainFrame(gClient->GetRoot(),850,650,kMainFrame | kHorizontalFrame);
   fMainFrame->SetLayoutManager(new TGTableLayout(fMainFrame,1,2,kFALSE,2));

//   fMainFrame->SetLayoutBroken(kFALSE);
   
   TGVerticalFrame* fSteeringFrame = new TGVerticalFrame(fMainFrame, 300, 600);


   fTab = new TGTab(fSteeringFrame, 300, 600);

   fTabContainer = fTab->AddTab("Draw");
   fTabContainer->ChangeOptions(kVerticalFrame);
   if (fViewerPar == 0x0) fViewerPar =  new MadxViewerPar(this);
   fViewerPar->Create(fTabContainer);
   
   fTabContainer = fTab->AddTab("Draw at");
   fTabContainer->ChangeOptions(kVerticalFrame);
   if (fViewerParAtEl == 0x0) fViewerParAtEl =  new MadxViewerParAtEl(this);
   fViewerParAtEl->Create(fTabContainer);

   fTabContainer = fTab->AddTab("Par Values");
   fTabContainer->ChangeOptions(kVerticalFrame);
   if (fParSetter == 0x0) fParSetter =  new MadxViewerParSetter(this);
   fParSetter->SetHostFrame(fTabContainer);
   
   
/************************************************************/


   // embedded canvas
   fParAtCanvas = new TRootEmbeddedCanvas("MADX Viewer",fMainFrame,600,600);
   Int_t wfParAtCanvas = fParAtCanvas->GetCanvasWindowId();
   
   TCanvas *c123 = new TCanvas("c123", 10, 10, wfParAtCanvas);
   fParAtCanvas->AdoptCanvas(c123);



/************************************************************/

   fSteeringFrame->AddFrame(fTab, new TGLayoutHints(kLHintsTop |kLHintsExpandY  | kLHintsFillY | kLHintsShrinkY ,2,2,2,2));
   fTab->MoveResize(2,2,300,600);
   
   
   TGCompositeFrame* fButtonsFrame = new TGCompositeFrame(fSteeringFrame);
   fButtonsFrame->SetLayoutManager(new TGTableLayout(fButtonsFrame,2,2,kTRUE,2));

   TGTableLayoutHints* tlh;

   fClearButton = new TGTextButton(fButtonsFrame,"Clear");
   fClearButton->SetTextJustify(36);
   fClearButton->Resize(115,40);
   tlh = new TGTableLayoutHints(0,1,0,1, kLHintsFillX | kLHintsExpandX ,12,5,5,5);
   fButtonsFrame->AddFrame(fClearButton, tlh);
   fClearButton->MoveResize(30,0,120,40);
   
   fExitButton = new TGTextButton(fButtonsFrame,"Exit");
   fExitButton->SetTextJustify(36);
   fExitButton->Resize(115,40);
   tlh = new TGTableLayoutHints(1,2,0,1, kLHintsFillX | kLHintsExpandX ,5,12,5,5);
   fButtonsFrame->AddFrame(fExitButton, tlh);
   fExitButton->MoveResize(30,0,120,40);


   fSaveAsButton = new TGTextButton(fButtonsFrame,"SaveAs");
   fSaveAsButton->SetTextJustify(36);
   fSaveAsButton->Resize(115,40);
   tlh = new TGTableLayoutHints(0,1,1,2, kLHintsFillX | kLHintsExpandX ,12,5,5,5);
   fButtonsFrame->AddFrame(fSaveAsButton, tlh);
   fSaveAsButton->MoveResize(30,0,120,40);


   fFormatButton = new TGTextButton(fButtonsFrame,"Format");
   fFormatButton->SetTextJustify(36);
   fFormatButton->Resize(115,40);
   tlh = new TGTableLayoutHints(1,2,1,2, kLHintsFillX |  kLHintsExpandX ,5,12,5,5);
   fButtonsFrame->AddFrame(fFormatButton, tlh);
   fFormatButton->MoveResize(30,0,120,40);



   fSteeringFrame->AddFrame(fButtonsFrame, new TGLayoutHints(kLHintsBottom | kLHintsShrinkY | kLHintsFillX | kLHintsExpandX, 2,2,2,2));
   fButtonsFrame->MoveResize(2,200,300,200);


/************************************************************/
/************************************************************/
   tlh = new TGTableLayoutHints(0,1,0,1, kLHintsExpandY  | kLHintsFillY | kLHintsShrinkY ,2,2,2,2);
   fMainFrame->AddFrame(fSteeringFrame, tlh);
   fSteeringFrame->MoveResize(2,2,300,600);
   
   
   
   fMainFrame->AddFrame(fParAtCanvas, new TGTableLayoutHints(1,2,0,1, 
                        kLHintsExpandX | kLHintsExpandY | kLHintsFillX | kLHintsFillY | kLHintsShrinkX | kLHintsShrinkY,
	                  2,2,2,2));


/************************************************************/
/************************************************************/

   fMainFrame->MapSubwindows();
   fMainFrame->Resize(fMainFrame->GetDefaultSize());
   fMainFrame->MapWindow();
   fMainFrame->MoveResize(10,10,900,600);
   fMainFrame->DontCallClose(); /*when we catch the this we will delete is ourselves*/
 
   fGEditor = new TGedEditor(c123);
   fGEditor->SetWMPosition(0, 800);
   fGEditor->Hide();
   
   
}
//______________________________________________________________________________________

void MadxViewer::Close()
{
//method that is called on close
  
  Info("Close","Executing Close, Terminating the application");
  Info("Close","Main win is %#x",fMainFrame);

  fMainFrame->DontCallClose(); /*when we catch the this we will delete is ourselves*/

  gApplication->SetReturnFromRun(kTRUE);
  gApplication->Terminate();

  Info("Close","Done");
}
//______________________________________________________________________________

void MadxViewer::ConnectSignals2Slots()
{
   // Connect check buttons.

   fMainFrame->Connect("CloseWindow()", "MadxViewer", this,"Close()");

   fExitButton->Connect("Clicked()", "MadxViewer", this,"Close()");

   fClearButton->Connect("Clicked()", "MadxViewer", this,"Clear()");
   
   fFormatButton->Connect("Clicked()","TGedEditor",fGEditor,"Show()");

   fSaveAsButton->Connect("Clicked()","MadxViewer",this,"SaveAs()");

   
   fViewerPar->ConnectSignals2Slots();
   fViewerParAtEl->ConnectSignals2Slots();
   fParSetter->ConnectSignals2Slots();
}
//______________________________________________________________________________

void MadxViewer::SaveAs()
{
  Info("SaveAs","");

  TCanvas* c = fParAtCanvas->GetCanvas();
  if ( c == 0x0)
   {
     Error("FocusOnCanvas","Canvas is NULL!");
     return;
   }


  static TString dir(".");
  static Int_t typeidx = 0;
  static Bool_t overwr = kFALSE;
  TGFileInfo fi;
  fi.fFileTypes   = fgkSaveAsTypes;
	                  
  fi.fIniDir      = StrDup(dir);
  fi.fFileTypeIdx = typeidx;
  fi.fOverwrite = overwr;
  new TGFileDialog(gClient->GetDefaultRoot(), fMainFrame, kFDSave, &fi);
  if (!fi.fFilename) return;
  Bool_t  appendedType = kFALSE;
  TString fn = fi.fFilename;
  TString ft = fi.fFileTypes[fi.fFileTypeIdx+1];
  dir     = fi.fIniDir;
  typeidx = fi.fFileTypeIdx;
  overwr  = fi.fOverwrite;
again:
  if (fn.EndsWith(".root") ||
      fn.EndsWith(".ps")   ||
      fn.EndsWith(".eps")  ||
      fn.EndsWith(".pdf")  ||
      fn.EndsWith(".svg")  ||
      fn.EndsWith(".gif")  ||
      fn.EndsWith(".xml")  ||
      fn.EndsWith(".xpm")  ||
      fn.EndsWith(".jpg")  ||
      fn.EndsWith(".png")  ||
      fn.EndsWith(".xcf")  ||
      fn.EndsWith(".tiff")) {
     c->SaveAs(fn);
  } else if (fn.EndsWith(".C"))
     c->SaveSource(fn);
  else {
     if (!appendedType) {
        if (ft.Index(".") != kNPOS) {
           fn += ft(ft.Index("."), ft.Length());
           appendedType = kTRUE;
           goto again;
        }
     }
     Warning("ProcessMessage", "file %s cannot be saved with this extension", fi.fFilename);
  }

  
}
//______________________________________________________________________________

void MadxViewer::Clear()
{
  Info("Clear","Cleaning");

  if (fActiveAtEl)
   {
     fViewerParAtEl->Clear();
   }
  else
   { 
     fViewerPar->Clear();
   }  

  ClearCanvas(); 
  UpdateCanvas();
}

//______________________________________________________________________________


void MadxViewer::UpdatePlots()
{
  Info("UpdatePlots","Updating plots");
  
  if (fActiveAtEl)
   {
     fViewerParAtEl->UpdatePlots();
   }
  else
   { 
     fViewerPar->UpdatePlots();
   }  
  
  UpdateCanvas();
  
}
//______________________________________________________________________________
void MadxViewer::HideKnobs(Int_t k1,Int_t k2)
{ 
  Info("HideKnobs","%d %d",k1,k2);
  if(fParSetter) fParSetter->HideKnobs(k1,k2);
}

//______________________________________________________________________________

void MadxViewer::ReadParametrs()
{

  Int_t nk = w_ptc_getnknobs();
  
  Info("ReadParametrs","Got %d knobs",nk);
  
  
  w_ptc_getknobsnames(); // sets names of the knobs in the comboboxes
  w_ptc_getfctnsnames(); // sets names of the available function in the combobox
  
  restart_sequ();
  Bool_t nextel = kFALSE;
  Int_t elno = 0;
  do
   { 
     elno++;
     TString ename(elno);
     ename += getcurrentelementname();
     Info("ReadParametrs","Element %d named %s",elno,ename.Data());
     
     fViewerParAtEl->AddElement(elno, ename);
     
     nextel = advance_node();
   }
  while( nextel != kFALSE);
  
     
}
//______________________________________________________________________________________

void MadxViewer::SetKnobName(Int_t n, const char* name)
{
  //Sets n'th knob name
  Info("SetKnobName","Function %d is %s",n,name);
  if (fViewerParAtEl) fViewerParAtEl->AddKnob(n,name);

  if (fParSetter) fParSetter->AddKnob(n,name);
  
}
//______________________________________________________________________________________

void MadxViewer::SetFunctionName(Int_t n, const char* name)
{
  //Sets n'th function name
  Info("SetFunctionName","Function %d is named %s",n,name);
  if (fViewerPar) fViewerPar->AddFunction(n, name);
  if (fViewerParAtEl) fViewerParAtEl->AddFunction(n, name);
}
//______________________________________________________________________________________


void MadxViewer::SetFunctionAt(Int_t el, Int_t n,  const char* name)
{
  //Sets n'th variable name
  if (fViewerParAtEl) fViewerParAtEl->SetFunctionAt(el, n, name);

}
//______________________________________________________________________________________

void MadxViewer::FocusOnCanvas()
{

  if ( fParAtCanvas == 0x0)
   {
     Error("FocusOnCanvas","Embedded Canvas is NULL!");
     return;
   }

  TCanvas* c = fParAtCanvas->GetCanvas();
  if ( c == 0x0)
   {
     Error("FocusOnCanvas","Canvas is NULL!");
     return;
   }
   
  c->cd();
  
}
//______________________________________________________________________________________

void MadxViewer::UpdateCanvas()
{

  if ( fParAtCanvas == 0x0)
   {
     Error("UpdateCanvas","Embedded Canvas is NULL!");
     return;
   }

  TCanvas* c = fParAtCanvas->GetCanvas();
  if ( c == 0x0)
   {
     Error("UpdateCanvas","Canvas is NULL!");
     return;
   }
  c->Modified();
  c->Update();
}

//______________________________________________________________________________________

void MadxViewer::ClearCanvas()
{

  if ( fParAtCanvas == 0x0)
   {
     Error("UpdateCanvas","Embedded Canvas is NULL!");
     return;
   }

  TCanvas* c = fParAtCanvas->GetCanvas();
  if ( c == 0x0)
   {
     Error("UpdateCanvas","Canvas is NULL!");
     return;
   }
  c->Clear();
}


//______________________________________________________________________________________

void MadxViewer::SetActiveAtEl()
{
  if (fActiveAtEl == kTRUE) return;

  ClearCanvas();
//  fParSetter->ZeroParameterValues();
  fViewerParAtEl->UpdatePlots();
  fViewerParAtEl->ReDraw();
  
  
  fActiveAtEl  = kTRUE;
}

//______________________________________________________________________________________

void MadxViewer::SetActiveAlong()
{
  if (fActiveAtEl == kFALSE) return;
  
  
  ClearCanvas();
  fViewerPar->UpdatePlots();
  fViewerPar->ReDraw();
  
  fActiveAtEl = kFALSE;
}
//______________________________________________________________________________________

Double_t MadxViewer::GetMadxFunction(Int_t n, Double_t *x, Double_t *par)
{
  if (fViewerParAtEl == 0x0)
   {
     Error("GetMadxFunction","fViewerParAtEl is NULL");
     return DBL_MAX;
   }
  
  return fViewerParAtEl->GetMadxFunction(n,x,par);
}


//______________________________________________________________________________________
//______________________________________________________________________________________
//______________________________________________________________________________________
//______________________________________________________________________________________
//______________________________________________________________________________________
//______________________________________________________________________________________


void runrviewer()
 {
   MadxViewer::Instance();
 }

extern "C" void madxviewer_setknobname(int* n, const char* name)
{
  MadxViewer::Instance()->SetKnobName(*n,name);
}

extern "C" void madxviewer_setfctnname(int* n, const char* name)
{
  MadxViewer::Instance()->SetFunctionName(*n,name);
}

extern "C" void madxviewer_setfunctionat(int* el, int* n,  const char* name)
{
  MadxViewer::Instance()->SetFunctionAt(*el,*n,name);
}



Double_t madxfctn00(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction( 0,x,par);}
Double_t madxfctn01(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction( 1,x,par);}
Double_t madxfctn02(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction( 2,x,par);}
Double_t madxfctn03(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction( 3,x,par);}
Double_t madxfctn04(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction( 4,x,par);}
Double_t madxfctn05(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction( 5,x,par);}
Double_t madxfctn06(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction( 6,x,par);}
Double_t madxfctn07(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction( 7,x,par);}
Double_t madxfctn08(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction( 8,x,par);}
Double_t madxfctn09(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction( 9,x,par);}

Double_t madxfctn10(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(10,x,par);}
Double_t madxfctn11(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(11,x,par);}
Double_t madxfctn12(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(12,x,par);}
Double_t madxfctn13(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(13,x,par);}
Double_t madxfctn14(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(14,x,par);}
Double_t madxfctn15(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(15,x,par);}
Double_t madxfctn16(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(16,x,par);}
Double_t madxfctn17(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(17,x,par);}
Double_t madxfctn18(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(18,x,par);}
Double_t madxfctn19(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(19,x,par);}

Double_t madxfctn20(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(20,x,par);}
Double_t madxfctn21(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(21,x,par);}
Double_t madxfctn22(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(22,x,par);}
Double_t madxfctn23(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(23,x,par);}
Double_t madxfctn24(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(24,x,par);}
Double_t madxfctn25(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(25,x,par);}
Double_t madxfctn26(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(26,x,par);}
Double_t madxfctn27(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(27,x,par);}
Double_t madxfctn28(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(28,x,par);}
Double_t madxfctn29(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(29,x,par);}


Double_t madxfctn30(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(30,x,par);}
Double_t madxfctn31(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(31,x,par);}
Double_t madxfctn32(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(32,x,par);}
Double_t madxfctn33(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(33,x,par);}
Double_t madxfctn34(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(34,x,par);}
Double_t madxfctn35(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(35,x,par);}
Double_t madxfctn36(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(36,x,par);}
Double_t madxfctn37(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(37,x,par);}
Double_t madxfctn38(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(38,x,par);}
Double_t madxfctn39(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(39,x,par);}

Double_t madxfctn40(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(40,x,par);}
Double_t madxfctn41(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(41,x,par);}
Double_t madxfctn42(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(42,x,par);}
Double_t madxfctn43(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(43,x,par);}
Double_t madxfctn44(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(44,x,par);}
Double_t madxfctn45(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(45,x,par);}
Double_t madxfctn46(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(46,x,par);}
Double_t madxfctn47(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(47,x,par);}
Double_t madxfctn48(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(48,x,par);}
Double_t madxfctn49(Double_t *x, Double_t *par) { return MadxViewer::Instance()->GetMadxFunction(49,x,par);}




//______________________________________________________________________________________
//______________________________________________________________________________________
//______________________________________________________________________________________
//______________________________________________________________________________________
//______________________________________________________________________________________
//______________________________________________________________________________________


ClassImp(MadxViewerParSetter)


MadxViewerParSetter::MadxViewerParSetter(MadxViewer* master):
 fMaster(master),
 fHostFrame(0x0),
 fFrames(new TObjArray(3,1)),
 fEntries(new TObjArray(3,1))
{
 //ctor
}
//______________________________________________________________________________________

MadxViewerParSetter::~MadxViewerParSetter()
{
  if (fFrames)
   {
     for (Int_t i=1;i<fFrames->GetEntriesFast();i++)
      {
        TGCompositeFrame* frame = dynamic_cast<TGCompositeFrame*>(fFrames->At(i));
        if (frame == 0x0)
         {
           Error("~MadxViewerParSetter","Sth crazy: fFrame contains not a compostie frame");
           continue;
         }
        frame->Cleanup();
      }
    
    fFrames->Delete();  
   }
   
  delete fFrames;
  delete fEntries;
  
}
//______________________________________________________________________________
void MadxViewerParSetter::HideKnobs(Int_t k1,Int_t k2)
{
  if (fFrames == 0x0) return;
  
  Info("HideKnobs","There are %d knob frames",fFrames->GetEntriesFast());
  for (Int_t i=1;i<=fFrames->GetEntriesFast();i++)
   {
     TGCompositeFrame* frame = dynamic_cast<TGCompositeFrame*>(fFrames->At(i));
     if (frame == 0x0)
      {
        Error("~MadxViewerParSetter","Sth crazy: fFrame contains not a compostie frame");
        continue;
      }
     if ( (i == k1) || (i == k2) )
      {
        fHostFrame->HideFrame(frame);
      }
     else
      {
        fHostFrame->ShowFrame(frame);
        SetParameterValue(i); //to synchronize what is on GUI with fortran value
      }
     
   }

}

//______________________________________________________________________________

void MadxViewerParSetter::AddKnob(Int_t n, const char* name)
{

   Info("AddKnob","Adding knob %d %s", n, name);
   
   if (fHostFrame == 0x0)
    { 
      Error("AddKnob","fHostFrame is NULL");
      return;
    }

   TGHorizontalFrame* f  = new TGHorizontalFrame(fHostFrame,200,30,kHorizontalFrame);

   TGLabel* l = new TGLabel(f,name);
   l->SetTextJustify(36);
   f->AddFrame( l, new TGLayoutHints(kLHintsLeft | kLHintsTop,6,2));
   l->MoveResize(5,2,100,20);
   
   TGNumberEntry* ne = new  TGNumberEntry(f,0.0);
   f->AddFrame( ne, new TGLayoutHints(kLHintsRight | kLHintsTop));
   ne->MoveResize(5,122,70,20);
   
   
   Int_t fromup = 5;
   if (n == 1) fromup = 20;
   fHostFrame->AddFrame(f, new TGLayoutHints( kLHintsRight | kLHintsTop, 5,5,fromup,5));
   f->MoveResize(2,2+n,200,30);

   fFrames->AddAtAndExpand(f,n);
   fEntries->AddAtAndExpand(ne,n);
   

   Info("AddKnob","Added knob %d %s", n, name);

   
}
//______________________________________________________________________________

void MadxViewerParSetter::ConnectSignals2Slots()
{

   Info("ConnectSignals2Slots","There is %d pars",fEntries->GetEntries());
   
   for (Int_t i=1; i<=fEntries->GetEntriesFast();i++)
    {
      TGNumberEntry* ne = dynamic_cast<TGNumberEntry*>(fEntries->At(i));
      if (ne == 0x0)
       {
         Error("ConnectSignals2Slots","TGNumberEntry at %d is NULL",i);
         continue;
       }

      ne->Connect("ValueSet(Long_t)","MadxViewerParSetter", this, "SetParameterValues()");
    }

}
//______________________________________________________________________________

void MadxViewerParSetter::SetParameterValue(Int_t n)
{

  TGNumberEntry* ne = dynamic_cast<TGNumberEntry*>(fEntries->At(n));
  if (ne == 0x0)
   {
     Error("SetParameterValue","TGNumberEntry at %d is NULL",n);
     return;
   }

  Float_t v =  ne->GetNumber();
  w_ptc_setparvalue(&n,&v);
}
//______________________________________________________________________________

void MadxViewerParSetter::SetParameterValues()
{
  for (Int_t i=1; i<=fEntries->GetEntriesFast();i++)
   {
     SetParameterValue(i);
   }

  fMaster->UpdatePlots();
  
}
//______________________________________________________________________________

void MadxViewerParSetter::ZeroParameterValues()
{

  for (Int_t i=1; i<=fEntries->GetEntriesFast();i++)
   {
     TGNumberEntry* ne = dynamic_cast<TGNumberEntry*>(fEntries->At(i));
     if (ne == 0x0)
      {
        Error("SetParameterValue","TGNumberEntry at %d is NULL",i);
        return;
      }

     ne->SetNumber(0.0);
   }   
}


//______________________________________________________________________________
//______________________________________________________________________________
//______________________________________________________________________________
//______________________________________________________________________________
//______________________________________________________________________________
