#ifndef MADXVIEWER_H
#define MADXVIEWER_H
/*Piotr Skowronski, CERN*/

#include "TGedFrame.h"
#include "TLegend.h"


class TRootEmbeddedCanvas;
class MadxViewerPar;
class MadxViewerParAtEl;
class TGedEditor;
class MadxViewerParSetter;
class TGTextButton;


class MadxViewer : public TObject {

public:
   static MadxViewer* Instance();

   MadxViewer();
   virtual ~MadxViewer();
   
   Bool_t Init();      
    
   void Create();
   void Close();
   void Clear();
   void SaveAs();
   
   void SetKnobName(Int_t n, const char* name);
   void SetFunctionName(Int_t n, const char* name);
   void SetFunctionAt(Int_t n, Int_t e, const char* name);
  
   void FocusOnCanvas(); 
   void UpdateCanvas(); 
   void ClearCanvas(); 

   void UpdatePlots();

   void SetActiveAtEl();   
   void SetActiveAlong();
   
   void HideKnobs(Int_t k1=-1,Int_t k2=-1);


   Double_t GetMadxFunction(Int_t n, Double_t *x, Double_t *par);

private:

   
   TGMainFrame         *fMainFrame;
   TGTab               *fTab;
   TGCompositeFrame    *fTabContainer;//not to be deleted belongs to fTab

   TGVerticalFrame     *fSteeringFrame;//not to be deleted belongs to fTab

   
   TGCompositeFrame    *fButtonsFrame;
   
   TGTextButton        *fClearButton;
   TGTextButton        *fExitButton;
   TGTextButton        *fSaveAsButton;
   TGTextButton        *fFormatButton;
   
   
   
   TRootEmbeddedCanvas *fParAtCanvas;
   
   MadxViewerPar       *fViewerPar;
   MadxViewerParAtEl   *fViewerParAtEl;
   MadxViewerParSetter *fParSetter;
   
   TGedEditor          *fGEditor;

   Bool_t               fActiveAtEl; // flag marking that Drawing at element is active
   
   static const char   *fgkSaveAsTypes[];
   static MadxViewer   *fgMadxViewer;// singleton MAD-X Plotter


   void ConnectSignals2Slots();
   void ReadParametrs();
   
   MadxViewer(const MadxViewer &){}
   MadxViewer &operator = (const MadxViewer &){return *this;}
   

   
   
   
   ClassDef(MadxViewer, 0) //GUI for editing TGLViewer attributes
};
//______________________________________________________________________________________


class MadxViewerParSetter : public TObject {

public:
   MadxViewerParSetter(MadxViewer* master = 0x0);
   virtual ~MadxViewerParSetter();
   
   Bool_t Init(){return kTRUE;}
   
   void ConnectSignals2Slots();
   
   void AddKnob(Int_t n, const char* name);
   void SetInvisible(Int_t f1, Int_t f2 = 0){}
   void SetParameterValue(Int_t n);
   void SetParameterValues();
   void ZeroParameterValues();
   void TestCaller(Long_t v){Info("Test Caller","%d",v);}
   void TestSimpleCaller(){Info("Test Caller","Simple Caller Called");}
   void SetHostFrame(TGCompositeFrame* f){fHostFrame = f;}
   void HideKnobs(Int_t k1=-1,Int_t k2=-1);
private:
   MadxViewer          *fMaster;
   TGCompositeFrame    *fHostFrame; //not to be deleted belongs to tab
   
   TObjArray           *fFrames;
   TObjArray           *fEntries;

   ClassDef(MadxViewerParSetter, 0) //GUI for editing TGLViewer attributes
};
//______________________________________________________________________________________


Double_t madxfctn00(Double_t *x, Double_t *par);
Double_t madxfctn01(Double_t *x, Double_t *par);
Double_t madxfctn02(Double_t *x, Double_t *par);
Double_t madxfctn03(Double_t *x, Double_t *par);
Double_t madxfctn04(Double_t *x, Double_t *par);
Double_t madxfctn05(Double_t *x, Double_t *par);
Double_t madxfctn06(Double_t *x, Double_t *par);
Double_t madxfctn07(Double_t *x, Double_t *par);
Double_t madxfctn08(Double_t *x, Double_t *par);
Double_t madxfctn09(Double_t *x, Double_t *par);

Double_t madxfctn10(Double_t *x, Double_t *par);
Double_t madxfctn11(Double_t *x, Double_t *par);
Double_t madxfctn12(Double_t *x, Double_t *par);
Double_t madxfctn13(Double_t *x, Double_t *par);
Double_t madxfctn14(Double_t *x, Double_t *par);
Double_t madxfctn15(Double_t *x, Double_t *par);
Double_t madxfctn16(Double_t *x, Double_t *par);
Double_t madxfctn17(Double_t *x, Double_t *par);
Double_t madxfctn18(Double_t *x, Double_t *par);
Double_t madxfctn19(Double_t *x, Double_t *par);

Double_t madxfctn20(Double_t *x, Double_t *par);
Double_t madxfctn21(Double_t *x, Double_t *par);
Double_t madxfctn22(Double_t *x, Double_t *par);
Double_t madxfctn23(Double_t *x, Double_t *par);
Double_t madxfctn24(Double_t *x, Double_t *par);
Double_t madxfctn25(Double_t *x, Double_t *par);
Double_t madxfctn26(Double_t *x, Double_t *par);
Double_t madxfctn27(Double_t *x, Double_t *par);
Double_t madxfctn28(Double_t *x, Double_t *par);
Double_t madxfctn29(Double_t *x, Double_t *par);

Double_t madxfctn30(Double_t *x, Double_t *par);
Double_t madxfctn31(Double_t *x, Double_t *par);
Double_t madxfctn32(Double_t *x, Double_t *par);
Double_t madxfctn33(Double_t *x, Double_t *par);
Double_t madxfctn34(Double_t *x, Double_t *par);
Double_t madxfctn35(Double_t *x, Double_t *par);
Double_t madxfctn36(Double_t *x, Double_t *par);
Double_t madxfctn37(Double_t *x, Double_t *par);
Double_t madxfctn38(Double_t *x, Double_t *par);
Double_t madxfctn39(Double_t *x, Double_t *par);

Double_t madxfctn40(Double_t *x, Double_t *par);
Double_t madxfctn41(Double_t *x, Double_t *par);
Double_t madxfctn42(Double_t *x, Double_t *par);
Double_t madxfctn43(Double_t *x, Double_t *par);
Double_t madxfctn44(Double_t *x, Double_t *par);
Double_t madxfctn45(Double_t *x, Double_t *par);
Double_t madxfctn46(Double_t *x, Double_t *par);
Double_t madxfctn47(Double_t *x, Double_t *par);
Double_t madxfctn48(Double_t *x, Double_t *par);
Double_t madxfctn49(Double_t *x, Double_t *par);



#ifndef WIN32

# define w_ptc_getnknobs w_ptc_getnknobs_
# define w_ptc_getknobsnames w_ptc_getknobsnames_
# define w_ptc_getfctnsnames w_ptc_getfctnsnames_
# define w_ptc_getfunctionat w_ptc_getfunctionat_
# define w_ptc_getfunctionvalueat w_ptc_getfunctionvalueat_
# define w_ptc_getnelements w_ptc_getnelements_
# define w_ptc_getlengthat w_ptc_getlengthat_
# define w_ptc_setparvalue w_ptc_setparvalue_
# define getcurrentelementname getcurrentelementname_
# define advance_node advance_node_
# define restart_sequ restart_sequ_

# define type_ofCall

#else

# define w_ptc_getnknobs W_PTC_GETNKNOBS
# define w_ptc_getknobsnames W_PTC_GETKNOBSNAMES
# define w_ptc_getfctnsnames W_PTC_GETFCTNSNAMES
# define w_ptc_getfunctionat W_PTC_GETFUNCTIONAT
# define w_ptc_getfunctionvalueat W_PTC_GETFUNCTIONVALUEAT
# define w_ptc_getnelements W_PTC_GETNELEMENTS
# define w_ptc_getlengthat W_PTC_GETLENGTHAT
# define w_ptc_setparvalue W_PTC_SETPARVALUE
# define getcurrentelementname GETCURRENTELEMENTNAME
# define advance_node ADVANCE_NODE
# define restart_sequ RESTART_SEQU

# define type_ofCall  _stdcall

#endif

/*********************************************/

extern "C" int  type_ofCall  w_ptc_getnknobs();
extern "C" void type_ofCall  w_ptc_getknobsnames();
extern "C" void type_ofCall  w_ptc_getfctnsnames();

extern "C" void  type_ofCall w_ptc_getfunctionat(int* n,int* e);
extern "C" float type_ofCall w_ptc_getfunctionvalueat(int* n,int* e);

extern "C" int   type_ofCall w_ptc_getnelements();
extern "C" float type_ofCall w_ptc_getlengthat(int* n);
extern "C" void  type_ofCall w_ptc_setparvalue(int* n, float* v);


extern "C" const char* type_ofCall getcurrentelementname();
extern "C" int   type_ofCall advance_node();
extern "C" int   type_ofCall restart_sequ();

/*********************************************/
extern "C" void runrviewer();
extern "C" void madxviewer_setknobname(int* n, const char* name);
extern "C" void madxviewer_setfctnname(int* n, const char* name);
extern "C" void madxviewer_setfunctionat(int* el, int* n,  const char* name);

#endif
