#include "MadxViewerParAtEl.h"
/*Piotr Skowronski, CERN*/

#include "MadxViewer.h"

#include <TF2.h>
#include <TAxis.h>
#include <TList.h>
#include "TLegend.h"

#include "TGLabel.h"
#include "TGComboBox.h"
#include "TGNumberEntry.h"
#include "TGButton.h"

ClassImp(MadxViewerParAtEl)

const Int_t MadxViewerParAtEl::fgkComboWidth = 120;
const Int_t MadxViewerParAtEl::fgkKnobComboWidth = 180;


MadxViewerParAtEl::MadxViewerParAtEl(MadxViewer* master):
 fMaster(master),
 fFormulas(0x0),
 fDrawnFunctions(new TList()),
 fLegend(0x0),
 fSame(kFALSE),
 fKn(0x0),
 fKn2(0x0),
 fFn(0x0),
 fEn(0x0)
{
 //ctor
}
//______________________________________________________________________________

MadxViewerParAtEl::~MadxViewerParAtEl()
{
//dtor  

 Info("~MadxViewerParAtEl","Deleting");

  if (fDrawnFunctions) fDrawnFunctions->SetOwner(kTRUE);
  delete fDrawnFunctions;

 Info("~MadxViewerParAtEl"," 1 ");
  
  delete fLabel3;
  delete fLabel2;

  delete fKnobTwoLabel;
  delete fKnobTwoComboBox;
  delete fKnobTwoFrame;

  delete fKnobOneLabel;
  delete fKnobOneComboBox;
  delete fKnobOneFrame;


  delete fKnobOneMinLabel;
  delete fKnobOneMinEntry;
  delete fKnobOneMaxLabel;
  delete fKnobOneMaxEntry;
  delete fKnobOneMinMaxFrame;


  delete fKnobTwoMinLabel;
  delete fKnobTwoMinEntry;
  delete fKnobTwoMaxLabel;
  delete fKnobTwoMaxEntry;
  delete fKnobTwoMinMaxFrame;

  delete fDrawOptionLabel;
  delete fDrawOptionComboBox;
  delete fDrawOptionFrame;

  delete fVarFrame;
  delete fVarLabel;
  delete fVarComboBox;

  delete fAtElFrame;
  delete fAtElLabel;
  delete fAtElComboBox;

  delete fSameButton;
  delete fDrawButton;


 Info("~MadxViewerParAtEl"," 2 ");
  
  
  delete [] fKn;
  delete [] fKn2;
  delete [] fFn;
  delete [] fEn;

 Info("~MadxViewerParAtEl"," 3 ");

  if (fFormulas)
   {
     Info("~MadxViewerParAtEl"," 4 ");

     Int_t nelem = w_ptc_getnelements(); 
     for (Int_t i=0; i<nelem;i++)
      {
        delete [] fFormulas[i];
      }
     delete [] fFormulas;
   }  

 Info("~MadxViewerParAtEl"," END ");


}
//______________________________________________________________________________

void MadxViewerParAtEl::Create(TGCompositeFrame* hostframe)
{

  ////////////////////////////////////////////////////
  // Variable 
  // TGHorizontalFrame
   if (hostframe == 0x0) return;
   
   fHostFrame = hostframe;
   
   fVarFrame = new TGHorizontalFrame(fHostFrame,fgkKnobComboWidth,30,kHorizontalFrame);
   fVarLabel = new TGLabel(fVarFrame,"Variable");
   fVarLabel->SetTextJustify(36);
   fVarFrame->AddFrame(fVarLabel, new TGLayoutHints(kLHintsLeft | kLHintsTop,10,5,2,2));
   fVarLabel->MoveResize(10,40,50,20);

 
   
   // combo box
   fVarComboBox = new TGComboBox(fVarFrame,-1,kHorizontalFrame | kSunkenFrame | kDoubleBorder | kOwnBackground);
   fVarComboBox->Resize(fgkComboWidth,20);
   fVarComboBox->Select(-1);
 
   fVarFrame->AddFrame(fVarComboBox, new TGLayoutHints(kLHintsRight | kLHintsTop,5,10,2,2));
   fVarComboBox->MoveResize(10,40,fgkComboWidth,20);
    
   fHostFrame->AddFrame(fVarFrame, new TGLayoutHints( kLHintsRight | kLHintsTop,2,2,10,5));
   fVarFrame->MoveResize(2,2,fgkKnobComboWidth,30);
   


  ////////////////////////////////////////////////////
  // At Element
  // TGHorizontalFrame
 
   fAtElFrame = new TGHorizontalFrame(fHostFrame,fgkKnobComboWidth,30,kHorizontalFrame);
   fAtElLabel = new TGLabel(fAtElFrame,"At Element");
   fAtElLabel->SetTextJustify(36);
   fAtElFrame->AddFrame(fAtElLabel, new TGLayoutHints(kLHintsLeft | kLHintsTop,10,5,2,2));
   fAtElLabel->MoveResize(10,40,50,20);
   
   // combo box
   fAtElComboBox = new TGComboBox(fAtElFrame,-1,kHorizontalFrame | kSunkenFrame | kDoubleBorder | kOwnBackground);
   fAtElComboBox->Resize(fgkComboWidth,20);
   fAtElComboBox->Select(-1);

   fAtElFrame->AddFrame(fAtElComboBox, new TGLayoutHints(kLHintsRight | kLHintsTop,5,10,2,2));
   fAtElComboBox->MoveResize(10,40,fgkComboWidth,20);

   //_________________

   fHostFrame->AddFrame(fAtElFrame, new TGLayoutHints( kLHintsRight | kLHintsTop,2,2,5,10));
   fAtElFrame->MoveResize(2,2,fgkKnobComboWidth,30);
    
   

   fLabel2 = new TGLabel(fHostFrame,"in function of knob");
   fLabel2->SetTextJustify(36);
   fHostFrame->AddFrame(fLabel2, new TGLayoutHints(kLHintsCenterX | kLHintsTop,2,2,10,2));
   fLabel2->MoveResize(2,2,fgkKnobComboWidth,30);
   
  ////////////////////////////////////////////////////
  // Par One

   fKnobOneFrame = new TGHorizontalFrame(fHostFrame,fgkKnobComboWidth,30,kHorizontalFrame);
   fKnobOneLabel = 0x0;
/*
   fKnobOneLabel = new TGLabel(fKnobOneFrame,"Parameter");
   fKnobOneLabel->SetTextJustify(36);
   fKnobOneFrame->AddFrame(fKnobOneLabel, new TGLayoutHints(kLHintsLeft | kLHintsTop,10,5));
   fKnobOneLabel->MoveResize(10,60,50,20);
*/

   // combo box
   fKnobOneComboBox = new TGComboBox(fKnobOneFrame,-1,kHorizontalFrame | kSunkenFrame | kDoubleBorder | kOwnBackground);
   
   fKnobOneComboBox->Resize(fgkKnobComboWidth,20);
//   fKnobOneComboBox->AddEntry("None",0);
   fKnobOneComboBox->Select(-1);
   

   fKnobOneFrame->AddFrame(fKnobOneComboBox, new TGLayoutHints(kLHintsRight | kLHintsTop,5,10));
   fKnobOneComboBox->MoveResize(2,60,fgkKnobComboWidth,20);
    
   fHostFrame->AddFrame(fKnobOneFrame, new TGLayoutHints( kLHintsRight | kLHintsTop,2,2,10,5));
   fKnobOneFrame->MoveResize(2,60,fgkKnobComboWidth,30);


   fKnobOneMinMaxFrame  = new TGHorizontalFrame(fHostFrame,fgkKnobComboWidth,30,kHorizontalFrame);
   fKnobOneMinLabel = new TGLabel(fKnobOneMinMaxFrame,"Min");
   fKnobOneMinLabel->SetTextJustify(36);
   fKnobOneMinMaxFrame->AddFrame( fKnobOneMinLabel, new TGLayoutHints(kLHintsLeft | kLHintsTop,10,5));
   
   
   fKnobOneMinEntry = new  TGNumberEntry(fKnobOneMinMaxFrame,-0.01);
   fKnobOneMinMaxFrame->AddFrame( fKnobOneMinEntry, new TGLayoutHints(kLHintsLeft | kLHintsTop));
   
   
   fKnobOneMaxLabel = new TGLabel(fKnobOneMinMaxFrame,"Max");
   fKnobOneMaxLabel->SetTextJustify(36);
   fKnobOneMinMaxFrame->AddFrame( fKnobOneMaxLabel , new TGLayoutHints(kLHintsLeft | kLHintsTop,10,5));

   fKnobOneMaxEntry =  new  TGNumberEntry(fKnobOneMinMaxFrame,0.01);
   fKnobOneMinMaxFrame->AddFrame( fKnobOneMaxEntry, new TGLayoutHints(kLHintsLeft | kLHintsTop,2,10));


   fHostFrame->AddFrame(fKnobOneMinMaxFrame, new TGLayoutHints( kLHintsRight | kLHintsTop,2,2,5,5));
   fKnobOneMinMaxFrame->MoveResize(2,60,fgkKnobComboWidth,30);
   

   fLabel3 = new TGLabel(fHostFrame,"and knob");
   fLabel3->SetTextJustify(36);
   fHostFrame->AddFrame(fLabel3, new TGLayoutHints(kLHintsCenterX | kLHintsTop,2,2,10,2));
   fLabel3->MoveResize(2,2,fgkKnobComboWidth,30);

   ////////////////////////////////////////////////////
   // Par Two
   // combo box

   fKnobTwoFrame = new TGHorizontalFrame(fHostFrame,fgkKnobComboWidth,30,kHorizontalFrame);
   fKnobTwoLabel = 0x0;
/*
   fKnobTwoLabel = new TGLabel(fKnobTwoFrame,"Parameter");
   fKnobTwoLabel->SetTextJustify(36);
   fKnobTwoFrame->AddFrame(fKnobTwoLabel, new TGLayoutHints(kLHintsLeft | kLHintsTop,10,5,10,10));
   fKnobTwoLabel->MoveResize(10,60,50,20);
*/
   
   fKnobTwoComboBox = new TGComboBox(fKnobTwoFrame,-1,kHorizontalFrame | kSunkenFrame | kDoubleBorder | kOwnBackground);
   fKnobTwoComboBox->Resize(fgkKnobComboWidth,20);
   fKnobTwoComboBox->AddEntry("None",0);
   fKnobTwoComboBox->Select(-1);

   fKnobTwoFrame->AddFrame(fKnobTwoComboBox, new TGLayoutHints(kLHintsRight | kLHintsTop,5,10,10,10));
   fKnobTwoComboBox->MoveResize(2,2,fgkKnobComboWidth,20);
    
   fHostFrame->AddFrame(fKnobTwoFrame, new TGLayoutHints( kLHintsRight | kLHintsTop,2,2,2,2));
   fKnobTwoFrame->MoveResize(2,80,fgkKnobComboWidth,30);





   fKnobTwoMinMaxFrame  = new TGHorizontalFrame(fHostFrame,fgkKnobComboWidth,30,kHorizontalFrame);
   fKnobTwoMinLabel = new TGLabel(fKnobTwoMinMaxFrame,"Min");
   fKnobTwoMinLabel->SetTextJustify(36);
   fKnobTwoMinMaxFrame->AddFrame( fKnobTwoMinLabel, new TGLayoutHints(kLHintsLeft | kLHintsTop,10,5));
   
   
   fKnobTwoMinEntry = new  TGNumberEntry(fKnobTwoMinMaxFrame,-0.01);
   fKnobTwoMinMaxFrame->AddFrame( fKnobTwoMinEntry, new TGLayoutHints(kLHintsLeft | kLHintsTop));
   
   
   fKnobTwoMaxLabel = new TGLabel(fKnobTwoMinMaxFrame,"Max");
   fKnobTwoMaxLabel->SetTextJustify(36);
   fKnobTwoMinMaxFrame->AddFrame( fKnobTwoMaxLabel , new TGLayoutHints(kLHintsLeft | kLHintsTop,10,5));

   fKnobTwoMaxEntry =  new  TGNumberEntry(fKnobTwoMinMaxFrame,0.01);
   fKnobTwoMinMaxFrame->AddFrame( fKnobTwoMaxEntry, new TGLayoutHints(kLHintsLeft | kLHintsTop,2,10));


   fHostFrame->AddFrame(fKnobTwoMinMaxFrame, new TGLayoutHints( kLHintsRight | kLHintsTop,2,2,5,5));
   fKnobTwoMinMaxFrame->MoveResize(2,60,fgkKnobComboWidth,30);
   


   //_________________________________________________ 2D Draw Option

   fDrawOptionFrame = new TGHorizontalFrame(fHostFrame,fgkKnobComboWidth,30,kHorizontalFrame);
   fDrawOptionLabel = new TGLabel(fDrawOptionFrame,"2D Draw Option");
   fDrawOptionLabel->SetTextJustify(36);
   fDrawOptionFrame->AddFrame(fDrawOptionLabel, new TGLayoutHints(kLHintsLeft | kLHintsTop,10,5,10,10));
   fDrawOptionLabel->MoveResize(10,60,50,20);

   
   fDrawOptionComboBox = new TGComboBox(fDrawOptionFrame,-1,kHorizontalFrame | kSunkenFrame | kDoubleBorder | kOwnBackground);
   fDrawOptionComboBox->Resize(80,20);
   fDrawOptionComboBox->AddEntry("colz",0);
   fDrawOptionComboBox->AddEntry("surf",1);
   fDrawOptionComboBox->AddEntry("surf1z",2);
   fDrawOptionComboBox->AddEntry("surf2z",3);
   fDrawOptionComboBox->AddEntry("surf3z",4);
   fDrawOptionComboBox->AddEntry("surf4z",5);
   fDrawOptionComboBox->AddEntry("surf5z",6);

   fDrawOptionComboBox->AddEntry("lego",7);
   fDrawOptionComboBox->AddEntry("lego1",8);
   fDrawOptionComboBox->AddEntry("lego2z",9);

   fDrawOptionComboBox->AddEntry("contz",10);
   fDrawOptionComboBox->AddEntry("cont1z",11);
   fDrawOptionComboBox->AddEntry("cont2",12);
   fDrawOptionComboBox->AddEntry("cont3",13);
   fDrawOptionComboBox->AddEntry("cont4z",14);

   fDrawOptionComboBox->AddEntry("box",15);
   fDrawOptionComboBox->AddEntry("box1",16);
   fDrawOptionComboBox->AddEntry("arr",17);
   fDrawOptionComboBox->AddEntry("scat",18);



   fDrawOptionComboBox->Select(0);

   fDrawOptionFrame->AddFrame(fDrawOptionComboBox, new TGLayoutHints(kLHintsRight | kLHintsTop,5,10,10,10));
   fDrawOptionComboBox->MoveResize(2,60,80,20);
    
   fHostFrame->AddFrame(fDrawOptionFrame, new TGLayoutHints( kLHintsRight | kLHintsTop,2,2,2,2));
   fDrawOptionFrame->MoveResize(2,80,120,30);
   



  ////////////////////////////////////////////////////
  // Buttons
   

   fDrawButton = new TGTextButton(fHostFrame,"Draw");
   fDrawButton->SetTextJustify(36);
   fDrawButton->Resize(115,40);
   fHostFrame->AddFrame(fDrawButton, new TGLayoutHints(kLHintsLeft | kLHintsBottom | kLHintsExpandX ,10,10,5,5));
   fDrawButton->MoveResize(30,0,120,40);


   fSameButton = new TGCheckButton(fHostFrame,"Add to plot");
   fHostFrame->AddFrame(fSameButton, new TGLayoutHints(kLHintsLeft | kLHintsTop,2,2,2,2));
   fSameButton->MoveResize(40,288,105,24);
   
   
   fLegend = new TLegend(0.8,0.8,0.95,0.95);
   fLegend->SetFillColor(0);


}
//______________________________________________________________________________

Bool_t MadxViewerParAtEl::Init()
{

  Int_t nelem = w_ptc_getnelements(); 
  
  if (nelem <= 0)
   {
     Error("Init","PTC reports no elements");
     return kFALSE;
   }

  if (fMaster == 0x0)
   {
     Error("Init","Master Viewer is not set");
     return kFALSE;
   }
  
  if (fVarComboBox == 0x0)
   {
     Error("Init","Parametric Viewer at Elements seems to be not created correctly");
     return kFALSE;
   }

  if ( fVarComboBox->GetNumberOfEntries() <= 0)
   {
     Error("Init","No parametric functions to show");
     return kFALSE;
   }
  fVarComboBox->Select(1);

  if ( fAtElComboBox->GetNumberOfEntries() <= 0)
   {
     Error("Init","No elements in the line");
     return kFALSE;
   }
  fAtElComboBox->Select(fAtElComboBox->GetNumberOfEntries() + 0 );
  
  
  if (fKnobOneComboBox->GetNumberOfEntries() > 0) 
   {
     fKnobOneComboBox->Select(1);
   }  
  
  if (fKnobTwoComboBox->GetNumberOfEntries() <= 2) 
   {
     fHostFrame->HideFrame(fKnobTwoFrame);
   } 
  else
   {
     fKnobTwoComboBox->Select(0);
   }

  
  fFormulas = new TString* [nelem];
  for (Int_t i=0; i<nelem;i++)
   {
     fFormulas[i] = new TString [fVarComboBox->GetNumberOfEntries()];
   }
  fKn = new Int_t[fVarComboBox->GetNumberOfEntries()]; //
  fKn2= new Int_t[fVarComboBox->GetNumberOfEntries()]; //
  fFn = new Int_t[fVarComboBox->GetNumberOfEntries()]; //to which (twiss) function it corresponds
  fEn = new Int_t[fVarComboBox->GetNumberOfEntries()]; //to which element it corresponds

  Info("Init","Kn2 is %#x",fKn2);
   
  
  return kTRUE;
}
//______________________________________________________________________________

void MadxViewerParAtEl::ConnectSignals2Slots()
{
   // Connect check buttons.
   if (fMaster == 0x0)
    {
      Error("ConnectSignals2Slots","Master Viewer is not set");
      return;
    }

   
   fDrawButton->Connect("Clicked()", "MadxViewerParAtEl", this,"DrawParametricAt()");
   fKnobTwoComboBox->Connect("Selected(Int_t)", "MadxViewerParAtEl", this,"AllowDrawSame()");
   fKnobOneComboBox->Connect("Selected(Int_t)", "MadxViewerParAtEl", this,"CheckKnobTwo()");
   fKnobOneMinEntry->Connect("ValueSet(Long_t)","MadxViewerParAtEl", this,"SetParOneRange()");
   fKnobOneMaxEntry->Connect("ValueSet(Long_t)","MadxViewerParAtEl", this,"SetParOneRange()");
   fKnobTwoMinEntry->Connect("ValueSet(Long_t)","MadxViewerParAtEl", this,"SetParTwoRange()");
   fKnobTwoMaxEntry->Connect("ValueSet(Long_t)","MadxViewerParAtEl", this,"SetParTwoRange()");

   fDrawOptionComboBox->Connect("Selected(Int_t)", "MadxViewerParAtEl", this,"SetDrawOption()");
   
}
//______________________________________________________________________________


void MadxViewerParAtEl::AllowDrawSame()
{
  
  if ( fKnobTwoComboBox->GetSelected() == fKnobOneComboBox->GetSelected() )
   {
     fKnobTwoComboBox->Select(0);
     return;
   }

  if ( fKnobTwoComboBox->GetSelected() > 0 )
   {
//     fSameButton->SetEnabled(kFALSE);
//     gStyle->SetOptTitle(kTRUE);
     fHostFrame->HideFrame(fSameButton);
     fHostFrame->ShowFrame(fKnobTwoMinMaxFrame);
     fHostFrame->ShowFrame(fDrawOptionFrame);

   }
  else
   {
     fHostFrame->ShowFrame(fSameButton);
     fHostFrame->HideFrame(fKnobTwoMinMaxFrame);
     fHostFrame->HideFrame(fDrawOptionFrame);
//     gStyle->SetOptTitle(kFALSE);
   }
  fHostFrame->Layout(); 
}
//______________________________________________________________________________
void MadxViewerParAtEl::CheckKnobTwo()
{
  if ( fKnobTwoComboBox->GetSelected() == fKnobOneComboBox->GetSelected() )
   {
     fKnobTwoComboBox->Select(0);
   }
}
//______________________________________________________________________________

void MadxViewerParAtEl::DrawParametricAt()
{
  
  Int_t linecolor = 1;
  
  if (fMaster == 0x0) return;
  
  
  fMaster->FocusOnCanvas();
  fMaster->SetActiveAtEl();
  
  fSelectedElement  = fAtElComboBox->GetSelected();
  fSelectedVar  = fVarComboBox->GetSelected();
  
  if (fSelectedParOne != fKnobOneComboBox->GetSelected())
   {
     fLegend->Clear();
     fDrawnFunctions->Delete();
     fSelectedParOne = fKnobOneComboBox->GetSelected();
      
   }
    
  if (fSelectedParTwo != fKnobTwoComboBox->GetSelected())
   {  
     fLegend->Clear();
     fDrawnFunctions->Delete();
     fSelectedParTwo = fKnobTwoComboBox->GetSelected();
   }
  

  fMaster->HideKnobs(fSelectedParOne,fSelectedParTwo);
  
   
  if(fFormulas[fSelectedElement-1][fSelectedVar-1].IsNull())
   {
     
//     Info("DrawParametricAt","Asking for element %d function %d ",fSelectedElement,fSelectedVar);
     w_ptc_getfunctionat(&fSelectedElement,&fSelectedVar);
     
     if(fFormulas[fSelectedElement-1][fSelectedVar-1].IsNull())
      {
        Error("DrawParametricAt","Function %d %d is empty",fSelectedElement,fSelectedVar);
        return;
      }
     
   }
  

  if (fSameButton->IsOn())
   {
     linecolor = fDrawnFunctions->GetEntries() + 1;
     if (fDrawnFunctions->GetEntries() > 9) //10 is white ....
      {
         linecolor++;
      }
   }
  else
   {
     fLegend->Clear();
     fDrawnFunctions->Delete();
   } 


  
  TString fn;
  fn = fVarComboBox->GetSelectedEntry()->GetTitle();
  fn+="@";
  fn+=fAtElComboBox->GetSelectedEntry()->GetTitle();
  
  if (fDrawnFunctions->FindObject(fn))
   {
     Info("DrawParametricAt","This function is already drawn");
     return;
   }
  
  if (fSelectedParTwo > 0)
   {
     fLegend->Clear();
     fDrawnFunctions->Delete();
   }
  
  TString formula(fFormulas[fSelectedElement-1][fSelectedVar-1]);
//  Info("DrawParametricAt","Original Formula: \n %s",formula.Data());

  TString xpar("p");
  if (fSelectedParOne < 10) xpar+="0";
  xpar += fSelectedParOne;
  
  if (formula.Contains(xpar))
   {
     formula.ReplaceAll(xpar+"^","x^");
     formula.ReplaceAll(xpar+"*","x*");
     formula.ReplaceAll(xpar+" ","x ");
   }
  else
   {
     formula += "+ 0.0*x";
   }   

//  Info("DrawParametricAt","X var substituted %s \n Formula: \n %s",xpar.Data(),formula.Data());
  
  
  Int_t counter=0;
  TString parname;
  TString parcode;
  for (Int_t i=1;i < fKnobOneComboBox->GetNumberOfEntries(); i++)
   {
     if (i == fSelectedParOne) continue;
     if (i == fSelectedParTwo) continue;
     
     
     parname = "p";
     if (i < 10) parname+= "0";
     parname+= i;
     parcode = "[";
     parcode+= counter;
     parcode+= "]";
     formula.ReplaceAll(parname+"^",parcode+"^");
     formula.ReplaceAll(parname+"*",parcode+"*");
     formula.ReplaceAll(parname+" ",parcode+" ");
     counter++;
   }

  formula.ReplaceAll("E-","*1^-");
  formula.ReplaceAll("E+","*1^+");

//  Info("DrawParametricAt","Final Formula: \n %s",formula.Data());
  
  Float_t xmin = fKnobOneMinEntry->GetNumber();
  Float_t xmax = fKnobOneMaxEntry->GetNumber();

  Int_t nf = fDrawnFunctions->GetEntries();
  
  TF1* f = 0x0;
  
  if (fSelectedParTwo > 0)
   {

     TString ypar("p");
     if (fSelectedParTwo < 10) ypar+="0";
     ypar += fSelectedParTwo;

     if (formula.Contains(ypar))
      {
       formula.ReplaceAll(ypar+"^","y^");
       formula.ReplaceAll(ypar+"*","y*");
       formula.ReplaceAll(ypar+" ","y ");
      }
     else
      {
        formula += "+ 0.0*y";
      }
      
     formula+=";"+xpar;
     formula+=";"+ypar;

     Float_t ymin = fKnobTwoMinEntry->GetNumber();
     Float_t ymax = fKnobTwoMaxEntry->GetNumber();
     
     
     fKn[nf] = fSelectedParOne;
     fKn2[nf]= fSelectedParTwo;
     fEn[nf] = fSelectedElement;
     fFn[nf] = fSelectedVar;
     
     Function_t cfctn = GetCFunction(nf);
     
     f = new TF2(fn, cfctn,xmin,xmax,ymin,ymax,0);
     
     f->GetXaxis()->SetTitle(xpar);
     f->GetYaxis()->SetTitle(ypar);
     
   }
  else
   {
     formula+=";"+xpar;
     
     fKn[nf] = fSelectedParOne;
     fEn[nf] = fSelectedElement;
     fFn[nf] = fSelectedVar;
     fKn2[nf] = -1;
     Function_t cfctn = GetCFunction(nf);
     f = new TF1(fn, cfctn, xmin,xmax,0);
     SetYRange(f);
     f->SetLineColor(linecolor);
     f->GetXaxis()->SetTitle(xpar);
   } 
  

  if (linecolor > 1)
   {
      f->Draw("same");
   }
  else 
   {
//     Info("DrawParametricAt","Deleting Existing Functions ... ");

     Info("DrawParametricAt","Drawing from scratch ... ");
     if (fSelectedParTwo > 0) 
      {
        Info("DrawParametricAt","Drawing 2D  ... ");
        f->Draw(fDrawOptionComboBox->GetSelectedEntry()->GetTitle());
 
        fLegend->SetX1NDC(0.02);
        fLegend->SetY1NDC(0.95);
        fLegend->SetX2NDC(0.98);
        fLegend->SetY2NDC(0.98);
        
      }
     else
      { 
        Info("DrawParametricAt","Drawing 1D  ... ");

        f->Draw();
        fLegend->SetX1NDC(0.8);
        fLegend->SetY1NDC(0.8);
        fLegend->SetX2NDC(0.95);
        fLegend->SetY2NDC(0.95);
      } 
   }


  fLegend->AddEntry(f,fn);
  fLegend->Draw();
  
  fDrawnFunctions->Add(f);
  
  
  fMaster->UpdateCanvas();
  
}
//______________________________________________________________________________
void MadxViewerParAtEl::UpdatePlots()
{
  SetYRange();
}
//______________________________________________________________________________

void MadxViewerParAtEl::ReDraw()
{
  
  Info("ReDraw","");
  
  Int_t n = 0;
  TIter next(fDrawnFunctions);
  TF1* ff = 0x0;
  
  while(( ff = (TF1*)next() ))
   { 
     if (n == 0)
      {
        ff->Draw();
      }
     else
      {
        ff->Draw("same");
      }
     n++; 
   }
  
  fLegend->Draw();
  fMaster->HideKnobs(fSelectedParOne,fSelectedParOne);
  fMaster->UpdateCanvas();  

}
//______________________________________________________________________________

void MadxViewerParAtEl::SetYRange(TF1* f)
 {
//Sets Y range in 1D functions

  if (fDrawnFunctions->GetEntries() == 0) return; //nothing to do
  
  if (f == 0x0) f=(TF1*)fDrawnFunctions->At(0);
  
  Double_t xmin = fKnobOneMinEntry->GetNumber();
  Double_t xmax = fKnobOneMaxEntry->GetNumber();

  Double_t min = f->GetMinimum(xmin,xmax);
  Double_t max = f->GetMaximum(xmin,xmax);

  Info("DrawParametricAt","This Fctn: %f %f ",min,max);

  TIter next(fDrawnFunctions);

  TF1* ff = 0x0;

  while(( ff = (TF1*)next() ))
   {

      Double_t cmax = ff->GetMaximum(xmin,xmax);

      if (cmax > max) 
       {
         max = cmax;
       }

      Double_t cmin = ff->GetMinimum(xmin,xmax);
      if (cmin < min) 
       {
         min = cmin;
       }

   }

  Info("DrawParametricAt","Global: %f %f ",min,max);

  Double_t delta = TMath::Max(TMath::Abs(min),TMath::Abs(max));

  min = min - delta*0.2;
  max = max + delta*0.2;

  Info("DrawParametricAt","New Limit: %f %f, delta %f",min,max,delta);
  ff = (TF1*)fDrawnFunctions->At(0);
  TAxis* yax = ff->GetYaxis();
  yax->SetRangeUser(min,max);
}
//______________________________________________________________________________

void MadxViewerParAtEl::SetParOneRange()
{
  
  Double_t min = fKnobOneMinEntry->GetNumber();
  Double_t max = fKnobOneMaxEntry->GetNumber();
  
  Info("SetParOneRange","Setting new X range %f %f",min,max);
  
  fKnobOneMinEntry->SetLimits(TGNumberFormat::kNELLimitMax,0,max);
  fKnobOneMaxEntry->SetLimits(TGNumberFormat::kNELLimitMin,min,0);
  
  TIter next(fDrawnFunctions);

  TF1* ff = 0x0;

  while(( ff = (TF1*)next() ))
   {
     ff->SetRange(min,max);
   }

  
  SetYRange();
  
  fMaster->UpdateCanvas();
  
}
//______________________________________________________________________________

void MadxViewerParAtEl::SetParTwoRange()
{
  
  Double_t ymin = fKnobTwoMinEntry->GetNumber();
  Double_t ymax = fKnobTwoMaxEntry->GetNumber();
  
  Info("SetParTwoRange","Setting new Y range %f %f",ymin,ymax);
  
  fKnobTwoMinEntry->SetLimits(TGNumberFormat::kNELLimitMax,0,ymax-0.000001);
  fKnobTwoMaxEntry->SetLimits(TGNumberFormat::kNELLimitMin,ymin+0.000001,0);

  TF2* ff = dynamic_cast<TF2*>(fDrawnFunctions->At(0));
  while(ff == 0x0) return;
  

  Double_t xmin = fKnobOneMinEntry->GetNumber();
  Double_t xmax = fKnobOneMaxEntry->GetNumber();


  ff->SetRange(xmin,ymin,xmax,ymax); 

  
  fMaster->UpdateCanvas();
  
}
//______________________________________________________________________________
void MadxViewerParAtEl::SetDrawOption()
{
  Info("SetDrawOption","");
  TF2* f = dynamic_cast<TF2*>(fDrawnFunctions->First());
  
  if (f == 0x0)
   {
     Error("SetDrawOption","Either there is not function drawn or it is not a 2D one");
     return;
   }
  f->SetDrawOption(fDrawOptionComboBox->GetSelectedEntry()->GetTitle());
  fMaster->UpdateCanvas();

}

//______________________________________________________________________________

void MadxViewerParAtEl::Clear()
{
  fLegend->Clear();
  fDrawnFunctions->Delete();
}
//______________________________________________________________________________

Double_t MadxViewerParAtEl::GetMadxFunction(Int_t n, Double_t *x, Double_t *par)
{
  Float_t v = x[0];
  Int_t knob = fKn[n]; // number of the knob that is variable of this function
//  Info("GetMadxFunction","n=%d, X axis is knob %d, value is %f",n,knob,v);
  w_ptc_setparvalue(&knob,&v);
  
  if (fKn2[n] > 0)
   { // two dimensional function
     knob = fKn2[n]; 
     v = x[1];
     w_ptc_setparvalue(&knob,&v);
   }

  Int_t j = fFn[n]; //function code
  Int_t i = fEn[n]; //element

  v =  w_ptc_getfunctionvalueat(&j,&i);

//  Info("GetMadxFunction","n=%d, Function code is %d at element %d,  value is %f",n,j,i,v);
  return v;
}
//______________________________________________________________________________
Function_t MadxViewerParAtEl::GetCFunction(Int_t n)
{

  switch(n)
   {
     case 0:
       return &madxfctn00;
     case 1:
       return &madxfctn01;
     case 2:
       return &madxfctn02;
     case 3:
       return &madxfctn03;
     case 4:
       return &madxfctn04;
     case 5:
       return &madxfctn05;
     case 6:
       return &madxfctn06;
     case 7:
       return &madxfctn07;
     case 8:
       return &madxfctn08;
     case 9:
       return &madxfctn09;

     case 10:
       return &madxfctn10;
     case 11:
       return &madxfctn11;
     case 12:
       return &madxfctn12;
     case 13:
       return &madxfctn13;
     case 14:
       return &madxfctn14;
     case 15:
       return &madxfctn15;
     case 16:
       return &madxfctn16;
     case 17:
       return &madxfctn17;
     case 18:
       return &madxfctn18;
     case 19:
       return &madxfctn19;

     case 20:
       return &madxfctn20;
     case 21:
       return &madxfctn21;
     case 22:
       return &madxfctn22;
     case 23:
       return &madxfctn23;
     case 24:
       return &madxfctn24;
     case 25:
       return &madxfctn25;
     case 26:
       return &madxfctn26;
     case 27:
       return &madxfctn27;
     case 28:
       return &madxfctn28;
     case 29:
       return &madxfctn29;

     case 30:
       return &madxfctn30;
     case 31:
       return &madxfctn31;
     case 32:
       return &madxfctn32;
     case 33:
       return &madxfctn33;
     case 34:
       return &madxfctn34;
     case 35:
       return &madxfctn35;
     case 36:
       return &madxfctn36;
     case 37:
       return &madxfctn37;
     case 38:
       return &madxfctn38;
     case 39:
       return &madxfctn39;

     case 40:
       return &madxfctn40;
     case 41:
       return &madxfctn41;
     case 42:
       return &madxfctn42;
     case 43:
       return &madxfctn43;
     case 44:
       return &madxfctn44;
     case 45:
       return &madxfctn45;
     case 46:
       return &madxfctn46;
     case 47:
       return &madxfctn47;
     case 48:
       return &madxfctn48;
     case 49:
       return &madxfctn49;
       
     default:
       Error("GetCFunction","There is not C Function corresponding to Variable %d",n);
       return 0x0;
   };
  
  return 0x0; 
  
}

//______________________________________________________________________________
void MadxViewerParAtEl::AddElement(Int_t n, const char* ename)
{
   if (fAtElComboBox) fAtElComboBox->AddEntry(ename,n);
}
//______________________________________________________________________________

void MadxViewerParAtEl::AddKnob(Int_t n, const char* name)
{
  if (fKnobOneComboBox) fKnobOneComboBox->AddEntry(name,n);
  if (fKnobTwoComboBox) fKnobTwoComboBox->AddEntry(name,n);

}
//______________________________________________________________________________

void MadxViewerParAtEl::AddFunction(Int_t n, const char* name )
{
   if (fVarComboBox) fVarComboBox->AddEntry(name,n);
}
//______________________________________________________________________________

void MadxViewerParAtEl::SetFunctionAt(Int_t el, Int_t n,  const char* name)
{
  //Sets n'th variable name
  Info("SetFunctionAt","Function %d at %d is %s",n,el,name);
  if (fFormulas ==  0x0)
   {
     Error("SetFunctionAt","fFormulas is NULL");
   }
  fFormulas[el-1][n-1]  = name;
  fFormulas[el-1][n-1].Append(" ");
}
