#ifndef MadxViewerParAtEl_H
#define MadxViewerParAtEl_H
/*Piotr Skowronski, CERN*/

#include "TGedFrame.h"

class MadxViewer;
class TGCheckButton;
class TGNumberEntry;
class TGButtonGroup;
class TGroupFrame;
class TGRadioButton;
class TGTabElement;
class TGButton;
class TGLViewer;
class TGTab;
class TGTabContainer;
class TGTextButton;
class TGComboBox;
class TString;
class TF1;
class TLegend;

typedef Double_t (*Function_t) (Double_t *, Double_t *);

class MadxViewerParAtEl : public TObject {

public:
   MadxViewerParAtEl(MadxViewer* master = 0x0);
   virtual ~MadxViewerParAtEl();
   
   void Create(TGCompositeFrame* hostframe);
   Bool_t Init();
   void ConnectSignals2Slots();

   void DrawParametricAt();
   void ReDraw();
   void UpdatePlots();
   void Clear();
   
   void AllowDrawSame();
   void CheckKnobTwo();
   
   void SetParOneRange();
   void SetParTwoRange();
   
   void SetDrawOption();
   
   void AddElement(Int_t n, const char* name);
   void AddKnob(Int_t n, const char* name);
   void AddFunction(Int_t n, const char* name);
   void SetFunctionAt(Int_t el, Int_t n,  const char* name);

   Double_t GetMadxFunction(Int_t n, Double_t *x, Double_t *par);
   
   Function_t GetCFunction(Int_t n);
private:
   
   MadxViewer          *fMaster;
   
   
   TGTextButton        *fDrawButton;
   
   TGCheckButton       *fSameButton;
   TGString            *fVarString;

   TGHorizontalFrame   *fVarFrame;
   TGLabel             *fVarLabel;
   TGComboBox          *fVarComboBox;

   TGHorizontalFrame   *fAtElFrame;
   TGLabel             *fAtElLabel;
   TGComboBox          *fAtElComboBox;

   TGHorizontalFrame   *fKnobOneFrame;
   TGLabel             *fKnobOneLabel;
   TGComboBox          *fKnobOneComboBox;

   TGHorizontalFrame   *fKnobOneMinMaxFrame;
   TGLabel             *fKnobOneMinLabel;
   TGNumberEntry       *fKnobOneMinEntry;
   TGLabel             *fKnobOneMaxLabel;
   TGNumberEntry       *fKnobOneMaxEntry;
   

   TGHorizontalFrame   *fKnobTwoMinMaxFrame;
   TGLabel             *fKnobTwoMinLabel;
   TGNumberEntry       *fKnobTwoMinEntry;
   TGLabel             *fKnobTwoMaxLabel;
   TGNumberEntry       *fKnobTwoMaxEntry;
   
   TGHorizontalFrame   *fKnobTwoFrame;
   TGLabel             *fKnobTwoLabel;
   TGComboBox          *fKnobTwoComboBox;

   TGHorizontalFrame   *fDrawOptionFrame;
   TGLabel             *fDrawOptionLabel;
   TGComboBox          *fDrawOptionComboBox;


   TGLabel             *fLabel2;
   TGLabel             *fLabel3;

   TGCompositeFrame    *fHostFrame; //not to be deleted belongs to tab

   TString            **fFormulas;
   
   TList               *fDrawnFunctions;
   TLegend             *fLegend;

   Bool_t fSame;
   Int_t  fSelectedElement;
   Int_t  fSelectedVar;
   Int_t  fSelectedParOne;
   Int_t  fSelectedParTwo;


   Int_t   *fKn;    //First variable: knob number 
   Int_t   *fKn2;   //Second variable: knob number 
   Int_t   *fFn;    //Function code, 1 is beta11, 2 is beta12, at the end are run-time user defined map elements 
   Int_t   *fEn;    //Element number

   static const Int_t fgkComboWidth;
   static const Int_t fgkKnobComboWidth;


   void SetYRange(TF1* f = 0x0);

   ClassDef(MadxViewerParAtEl, 0) //GUI for editing TGLViewer attributes
};


#endif
